# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2022, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)

"""
.. moduleauthors:: 
   Giovanni Rabuffo <giovanni.rabuffo@univ-amu.fr>, 
   Carmela Calabrese <carmela.calabrese@iit.it>,
   Jan Fousek <jan.fousek@univ-amu.fr>, 

   Under the "NextGen" Research Infrastructure Voucher SC3 associated to the HBP Flagship as a Partnering Project (PP)
   Project leader: Simona Olmi <simone.olmi@gmail.com>
   EBRAINS Partner: Viktor Jirsa <viktor.jirsa@univ-amu.fr>
"""

from tvb.simulator.models.base import Model
from tvb.basic.neotraits.api import NArray, List, Range, Final

import numpy

class KIonEx(Model):
    r"""
    KIonEx (Potassium K+ Ion exchange) mean-field model was developed in (Bandyopadhyay & Rabuffo et al. 2023). 
    It describes the mean-field activity of a population of Hodgkin-Huxley-type neurons (Depannemaker et al 2022) 
    linking the slow fluctuations of intra- and extra-cellular potassium ion concentrations to the mean membrane potential, 
    and the synaptic input to the population firing rate. 
    The model is derived as the mathematical limit of an infinite number of all-to-all coupled neurons, resulting in 5 state variables:
    :math:`x` represents a phenomenological variable connected to the firing rate, 
    :math:`V` represent the average membrane potential,
    :math:`n` represents the gating variable for potassium K, 
    :math:`\Delta K_{int}` represent the intracellular potassium concentration,
    :math:`K_g` represents the extracellular potassium buffering by the external bath
    """
    #_ui_name = "KIonEx"
    #ui_configurable_parameters = ['E', 'K_bath', 'J', 'eta', 'Delta','c_minus','R_minus','c_plus','R_plus','Vstar']

    E = NArray(
        label=r":math:`E`",
        default=numpy.array([0.]),
        domain=Range(lo=-80, hi=0, step=0.5),
        doc="""Reversal Potential""",
    )

    K_bath = NArray(
        label=r":math:`K_bath`",
        default=numpy.array([5.5]),
        domain=Range(lo=3, hi=40.0, step=0.25),
        doc="""Potassium concentration in bath""",
    )

    J = NArray(
        label=r":math:`J`",
        default=numpy.array([0.1]),
        domain=Range(lo=0.001, hi=40.0, step=0.01),
        doc="""Mean Synaptic weight""",
    )

    eta = NArray(
        label=":math:`eta`",
        default=numpy.array([0.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.1),
        doc="""Mean heterogeneous noise""",
    )

    Delta = NArray(
        label=r":math:`\Delta`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.0, hi=10.0, step=0.01),
        doc="""HWHM heterogeneous noise""",
    )

    c_minus = NArray(
        label=":math:`c_minus`",
        default=numpy.array([-40.0]),
        domain=Range(lo=-100.0, hi=-10.0, step=0.5),
        doc="""x-coordinate left parabola""",
    )

    R_minus = NArray(
        label=r":math:`R_minus`",
        default=numpy.array([0.5]),
        domain=Range(lo=0.0001, hi=5.0, step=0.01),
        doc="""curvature left parabola""",
    )

    c_plus = NArray(
        label=":math:`c_plus`",
        default=numpy.array([-20.0]),
        domain=Range(lo=-80.0, hi=0.0, step=0.5),
        doc="""x-coordinate right parabola""",
    )

    R_plus = NArray(
        label=r":math:`R_plus`",
        default=numpy.array([-0.5]),
        domain=Range(lo=-5.0, hi=-0.0001, step=0.01),
        doc="""curvature right parabola""",
    )


    Vstar = NArray(
        label=r":math:`Vstar`",
        default=numpy.array([-31]),
        domain=Range(lo=-55.0, hi=-15, step=0.5),
        doc="""x-coordinate meeting point of parabolas""",
    )

    #'Cm': 1, #nF, # membrane capacitance
    Cm = NArray(
        label=r":math:`Cm`",
        default=numpy.array([1]),
        domain=Range(lo=0.5, hi=1.5, step=0.1),
        doc="""membrane capacitance""",
    )

    # 'tau_n': 4, # ms # time constant of gating variable
    tau_n = NArray(
        label=r":math:`tau_n`",
        default=numpy.array([4]),
        domain=Range(lo=2, hi=6, step=0.5),
        doc="""time constant of gating variable""",
    )

    # 'gamma': 0.04,  # mol / C  # conversion factor
    gamma = NArray(
        label=r":math:`gamma`",
        default=numpy.array([0.04]),
        domain=Range(lo=0.02, hi=0.06, step=0.005),
        doc="""conversion factor""",
    )

    #'epsilon': 0.001, # mHz  # diffusion rate
    epsilon = NArray(
        label=r":math:`epsilon`",
        default=numpy.array([0.001]),
        domain=Range(lo=0.0005, hi=0.0015, step=0.0001),
        doc="""diffusion rate""",
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={
            "x": numpy.array([0.0, numpy.inf]),
            "V": numpy.array([-500.0, numpy.inf]),
            "n": numpy.array([0.0, numpy.inf]),
            "DKi": numpy.array([-100.0, numpy.inf]),
            "Kg": numpy.array([-100.0, numpy.inf])
        },
    )

    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={
            "x": numpy.array([0., 1]),
            "V": numpy.array([-90., 10.]),
            "n": numpy.array([0., 1]),
            "DKi": numpy.array([-10, 0]),
            "Kg": numpy.array([-20, -5])
        },
        doc="""Expected ranges of the state variables for initial condition generation and phase plane setup.""",
    )



    # TODO should match cvars below..
    coupling_terms = Final(
        label="Coupling terms",
        # how to unpack coupling array
        default=["Coupling_Term"]
    )

    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=("x", "V","n","DKi", "Kg"),
        default=("x", "V","n","DKi", "Kg"),
        doc="The quantities of interest for monitoring for the Infinite HH 5D.",
    )

    state_variables = ['x', 'V','n','DKi','Kg']
    _nvar = 5
    # Cvar is the coupling variable. 
    cvar = numpy.array([0], dtype=numpy.int32)
    # Stvar is the variable where stimulus is applied.
    stvar = numpy.array([1], dtype=numpy.int32)
    
    def dfun(self, state_variables, coupling, local_coupling=0.0):
        r"""
        The mean-field approximation for a population of Hodgkin-Huxley-type neurons driven by slow potassium dynamics consists of a 5D system:

        .. math::
            \frac{dx}{dt}&=
            \begin{cases} 
            \Delta+2R_{-}(V-c_{-})x - J r x; \  V\leq V^{\star}\\
            \Delta+2R_{+}(V-c_{+})x - J r x; \  V> V^{\star},
            \end{cases}\\
            \frac{dV}{dt}&=
            \begin{cases} 
            -\frac{1}{C_m}(I_{Cl}+I_{Na}+I_{K}+I_{pump})-R_{-}x^2+J r(E_{syn}-V)+\overline{\eta}; \  V\leq V^{\star}\\
            -\frac{1}{C_m}(I_{Cl}+I_{Na}+I_{K}+I_{pump})-R_{+}x^2+J r(E_{syn}-V)+\overline{\eta}; \  V>V^{\star}, 
            \end{cases}\\
            \frac{dn}{dt} &= \frac{n_{\infty}(V)-n}{\tau_n}, \\
            \frac{d \Delta [K^{+}]_{int}}{dt} &= - \frac{\gamma}{\omega_i}(I_K - 2 I_{pump}),\\
            \frac{d[K^+]_g}{dt} &= \epsilon ([K^+]_{bath} - [K^+]_{ext}\}).\\

        For details refer to (Bandyopadhyay & Rabuffo et al. 2023)
        """
        
        x = state_variables[0, :]
        V = state_variables[1, :]
        n = state_variables[2, :]
        DKi = state_variables[3, :]
        Kg = state_variables[4, :]
        
        #[State_variables, nodes]
        E = self.E

        K_bath = self.K_bath
        J = self.J
        eta = self.eta
        Delta = self.Delta

        c_minus = self.c_minus
        R_minus = self.R_minus
        c_plus = self.c_plus
        R_plus = self.R_plus
        Vstar = self.Vstar

        Cm      = self.Cm
        tau_n   = self.tau_n
        gamma   = self.gamma
        epsilon = self.epsilon
     
        Coupling_Term = coupling[0, :] #This zero refers to the first element of cvar (trivial in this case)

        # Constants
        Cnap = 21.0  # mol.m**-3 
        DCnap = 2.0  # mol.m**-3 
        Ckp = 5.5  # mol.m**-3 
        DCkp = 1.0  # mol.m**-3 
        Cmna = -24.0  # mV 
        DCmna = 12.0  # mV 
        Chn = 0.4  # dimensionless 
        DChn = -8.0  # dimensionless 
        Cnk = -19.0  # mV 
        DCnk = 18.0  # mV #Ok in the paper
        g_Cl = 7.5  # nS #Ok in the paper   # chloride conductance
        g_Na = 40.0  # nS   # maximal sodiumconductance
        g_K = 22.0  # nS  # maximal potassium conductance
        g_Nal = 0.02  # nS  # sodium leak conductance
        g_Kl = 0.12  # nS  # potassium leak conductance
        rho = 250.  # 250.,#pA # maximal Na/K pump current
        w_i = 2160.0  # umeter**3  # intracellular volume 
        w_o = 720.0  # umeter**3 # extracellular volume 
        Na_i0 = 16.0  # mMol/m**3 # initial concentration of intracellular Na
        Na_o0 = 138.0 # mMol/m**3 # initial concentration of extracellular Na
        K_i0 = 130.0  # mMol/m**3 # initial concentration of intracellular K
        K_o0 = 4.80   # mMol/m**3 # initial concentration of extracellular K
        Cl_i0 = 5.0   # mMol/m**3 # initial concentration of intracellular Cl
        Cl_o0 = 112.0 # mMol/m**3 # initial concentration of extracellular Cl
        

        # helper functions

        def m_inf(V):
            return 1.0/(1.0+numpy.exp((Cmna-V)/DCmna))

        def n_inf(V):
            return 1.0/(1.0+numpy.exp((Cnk-V)/DCnk))

        def h(n):
            return 1.1 - 1.0 / (1.0 + numpy.exp(-8.0 * (n - 0.4)))

        def I_K_form(V,n,K_o,K_i):
            return (g_Kl+g_K*n)*(V- 26.64*numpy.log(K_o/K_i)) 

        def I_Na_form(V,Na_o,Na_i,n):
            return (g_Nal+g_Na*m_inf(V)*h(n))*(V- 26.64*numpy.log(Na_o/Na_i))

        def I_Cl_form(V):
            return g_Cl*(V+ 26.64*numpy.log(Cl_o0/Cl_i0)) 

        def I_pump_form(Na_i,K_o):
            return rho*(1.0/(1.0+numpy.exp((Cnap - Na_i) / DCnap))*(1.0/(1.0+numpy.exp((Ckp - K_o)/DCkp)))) 

        def V_dot_form(I_Na,I_K,I_Cl,I_pump):
            return (-1.0/Cm)*(I_Na+I_K+I_Cl+I_pump) 

        beta= w_i / w_o 
        DNa_i = -DKi 
        DNa_o = -beta * DNa_i
        DK_o = -beta * DKi
        K_i = K_i0 + DKi 
        Na_i = Na_i0 + DNa_i 
        Na_o = Na_o0 + DNa_o 
        K_o = K_o0 + DK_o + Kg 

        ninf=n_inf(V)
        I_K = I_K_form(V,n,K_o,K_i)
        I_Na = I_Na_form(V,Na_o,Na_i,n)
        I_Cl = I_Cl_form(V)
        I_pump = I_pump_form(Na_i,K_o)
    
        r = R_minus*x/numpy.pi
        Vdot = (-1.0/Cm)*(I_Na+I_K+I_Cl+I_pump) 

        derivative = numpy.empty_like(state_variables)

        if_xdot = Delta+2*R_minus*(V-c_minus)*x-J*r*x 
        else_xdot = Delta+2*R_plus*(V-c_plus)*x-J*r*x
        derivative[0] = numpy.where(V <= Vstar, if_xdot, else_xdot)

        if_Vdot = Vdot - R_minus*x**2 + eta + (R_minus/numpy.pi)*Coupling_Term*(E-V)
        else_Vdot = Vdot - R_plus*x**2 + eta + (R_minus/numpy.pi)*Coupling_Term*(E-V)
        derivative[1] = numpy.where(V <= Vstar, if_Vdot, else_Vdot)

        derivative[2] = (ninf - n) / tau_n
        derivative[3] = -(gamma / w_i) * (I_K - 2.0 * I_pump)
        derivative[4] = epsilon * (K_bath - K_o)
        
        return derivative



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Test monitors for specific properties such as default periods, correct
handling of projection matrices, etc.

.. moduleauthor:: Paula Sanz Leon <paula@tvb.invalid>
.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""

import numpy
from tvb.datatypes.surfaces import CorticalSurface
from tvb.tests.library.base_testcase import BaseTestCase
from tvb.datatypes import sensors
from tvb.simulator import monitors, models, coupling, integrators, noise, simulator
from tvb.datatypes import connectivity
from tvb.datatypes.cortex import Cortex
from tvb.datatypes.region_mapping import RegionMapping
from tvb.datatypes.projections import ProjectionSurfaceEEG
from tvb.datatypes.sensors import SensorsInternal


class TestMonitors(BaseTestCase):
    """
    Define test cases for monitors:
        - initialise each class
        - check default parameters (period)
        -
    """

    default_period = 0.9765625  # 1024Hz

    def test_monitor_raw(self):
        monitors.Raw()

    def test_monitor_rawvoi(self):
        monitors.RawVoi()

    def test_monitor_tavg(self):
        monitor = monitors.TemporalAverage()
        assert monitor.period == self.default_period

    def test_monitor_afferentcoupling(self):
        monitors.AfferentCoupling()

    def test_monitor_afferentcouplingtavg(self):
        monitors.AfferentCouplingTemporalAverage()

    def test_monitor_gavg(self):
        monitor = monitors.GlobalAverage()
        assert monitor.period == self.default_period

    def test_monitor_savg(self):
        monitor = monitors.SpatialAverage()
        assert monitor.period == self.default_period

    def test_monitor_subsample(self):
        monitor = monitors.SubSample()
        assert monitor.period == self.default_period

    def test_monitor_eeg(self):
        monitor = monitors.EEG()
        assert monitor.period == self.default_period

    def test_monitor_meg(self):
        monitor = monitors.MEG()
        assert monitor.period == self.default_period

    def test_monitor_stereoeeg(self):
        """
        This has to be verified.
        """
        monitor = monitors.iEEG()
        monitor.sensors = sensors.SensorsInternal.from_file()
        assert monitor.period == self.default_period

    def test_monitor_bold(self):
        """
        This has to be verified.
        """
        monitor = monitors.Bold()
        assert monitor.period == 2000.0


class TestMonitorsConfiguration(BaseTestCase):
    """
    Configure Monitors

    """

    def test_monitor_bold(self):
        """
        This has to be verified.
        """
        monitor = monitors.Bold()
        assert monitor.period == 2000.0


class TestProjectionMonitorsWithSubcorticalRegions(BaseTestCase):
    """
    Cortical surface with subcortical regions, sEEG, EEG & MEG, using a stochastic
    integration. This test verifies the shapes of the projection matrices, and
    indirectly covers region mapping, etc.

    """

    # hard code parameters to smoke test
    speed = 4.0
    period = 1e3 / 1024.0  # 1024 Hz
    coupling_a = numpy.array([0.014])
    n_regions = 192

    def test_surface_sim_with_projections(self):

        # Setup Simulator obj
        oscillator = models.Generic2dOscillator()
        white_matter = connectivity.Connectivity.from_file('connectivity_%d.zip' % (self.n_regions,))
        white_matter.speed = numpy.array([self.speed])
        white_matter_coupling = coupling.Difference(a=self.coupling_a)
        heunint = integrators.HeunStochastic(
            dt=2 ** -4,
            noise=noise.Additive(nsig=numpy.array([2 ** -10, ]))
        )
        mons = (
            monitors.EEG.from_file(period=self.period),
            monitors.MEG.from_file(period=self.period),
            # monitors.iEEG.from_file(period=self.period),
            # SEEG projection data is not part of tvb-data on Pypi, thus this can not work generic
        )
        local_coupling_strength = numpy.array([2 ** -10])
        region_mapping = RegionMapping.from_file('regionMapping_16k_%d.txt' % (self.n_regions,))
        region_mapping.surface = CorticalSurface.from_file()
        default_cortex = Cortex.from_file()
        default_cortex.region_mapping_data = region_mapping
        default_cortex.coupling_strength = local_coupling_strength
        default_cortex.region_mapping_data.connectivity = white_matter

        sim = simulator.Simulator(model=oscillator, connectivity=white_matter, coupling=white_matter_coupling,
                                  integrator=heunint, monitors=mons, surface=default_cortex)

        with numpy.errstate(all='ignore'):
            sim.configure()

        # check configured simulation connectivity attribute
        conn = sim.connectivity
        assert conn.number_of_regions == self.n_regions
        assert conn.speed == self.speed

        # test monitor properties
        lc_n_node = sim.surface.local_connectivity.matrix.shape[0]
        for mon in sim.monitors:
            assert mon.period == self.period
            n_sens, g_n_node = mon.gain.shape
            assert g_n_node == sim.number_of_nodes
            assert n_sens == mon.sensors.number_of_sensors
            assert lc_n_node == g_n_node

        # check output shape
        ys = {}
        mons = 'eeg meg seeg'.split()
        for key in mons:
            ys[key] = []
        for data in sim(simulation_length=3.0):
            for key, dat in zip(mons, data):
                if dat:
                    _, y = dat
                    ys[key].append(y)
        for mon, key in zip(sim.monitors, mons):
            ys[key] = numpy.array(ys[key])
            assert ys[key].shape[2] == mon.gain.shape[0]


class TestProjectionMonitorsWithNoSubcortical(TestProjectionMonitorsWithSubcorticalRegions):
    """
    Idem. but with the 76 regions connectivity, where no sub-cortical regions are included
    """
    n_regions = 76


class TestAllAnalyticWithSubcortical(BaseTestCase):
    """Test correct gain matrix shape for all analytic with subcortical nodes."""

    def _build_test_sim(self):
        sim = simulator.Simulator(
            connectivity=connectivity.Connectivity.from_file('connectivity_192.zip'),
            monitors=(monitors.iEEG(
                sensors=SensorsInternal.from_file(),
                region_mapping=RegionMapping.from_file('regionMapping_16k_192.txt')
            ),)
        ).configure()
        return sim

    def _build_test_sim_eeg(self):
        eeg_monitor = monitors.EEG()
        eeg_monitor.sensors = sensors.SensorsEEG.from_file()
        eeg_monitor.region_mapping = RegionMapping.from_file('regionMapping_16k_192.txt')
        eeg_monitor.projection = ProjectionSurfaceEEG.from_file()

        sim = simulator.Simulator(
                connectivity=connectivity.Connectivity.from_file('connectivity_192.zip'),
                monitors=[eeg_monitor],
                simulation_length=100
        ).configure()
        return sim

    def test_gain_size(self):
        sim = self._build_test_sim()
        ieeg = sim.monitors[0]  # type: SensorsInternal
        n_sens, n_reg = ieeg.gain.shape
        assert ieeg.sensors.locations.shape[0] == n_sens
        assert sim.connectivity.number_of_regions == n_reg

    def test_gain_config_idempotent(self):
        "Check that rerunning the config doesn't increase matrix size"
        sim = self._build_test_sim_eeg()
        eeg, = sim.monitors
        eeg: monitors.EEG
        initial_gain_shape = eeg.gain.shape
        eeg.config_for_sim(sim)
        reconfig_gain_shape = eeg.gain.shape
        assert initial_gain_shape == reconfig_gain_shape
        eeg.config_for_sim(sim)
        rereconfig_gain_shape = eeg.gain.shape
        assert reconfig_gain_shape == rereconfig_gain_shape

    def test_gain_order(self):
        conn = connectivity.Connectivity()
        conn.generate_surrogate_connectivity(4)
        conn.centres /= 100.0
        conn.cortical = numpy.array([1, 0, 1, 1], numpy.bool_)
        seeg_sensors = SensorsInternal(
            locations=conn.centres,
            labels=conn.region_labels)
        region_mapping = RegionMapping(
            array_data=numpy.r_[:conn.number_of_regions],
            connectivity=conn)
        seeg_monitor = monitors.iEEG(
            sensors=seeg_sensors,
            region_mapping=region_mapping,
            )
        sim = simulator.Simulator(
            connectivity=conn,
            monitors=[seeg_monitor],
        )
        sim.configure()
        # NB: each sensor above is on one node, in same order
        # they must create NaN values in gain matrix
        # we zero NaNs
        # order of zeros by tell us order of node-sensor pairs
        zero_mask = seeg_monitor.gain == 0.0
        row, col = numpy.where(zero_mask)
        assert (row == col).all()


class TestSVEEG(BaseTestCase):
    "Test use of multiple state variables in monitors."

    def _build_test_sim_eeg(self):
        import numpy as np

        conn = connectivity.Connectivity.from_file()
        conn.speed = np.r_[70.0]
        cfun = coupling.Linear(a=np.r_[0.2], b = np.r_[0.0])

        eeg = monitors.EEG.from_file()
        eeg.period = 1.0
        eeg.variables_of_interest = np.r_[0]
        sim = simulator.Simulator(
            model=models.ReducedSetHindmarshRose(),
            integrator=integrators.HeunDeterministic(dt=0.0122),
            connectivity=conn,
            coupling=cfun,
            monitors=(eeg,),
            simulation_length=10.0
        )
        return sim

    def test_config_run(self):
        sim = self._build_test_sim_eeg()
        sim.configure()
        (_, d),  = sim.run()
        assert d.shape[2] == 65



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""

The Spectral datatypes.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""
import numpy

from tvb.basic.neotraits.api import HasTraits, Attr, NArray, Int, Float, EnumAttr, TVBEnum
from tvb.datatypes import time_series


class WindowingFunctionsEnum(TVBEnum):
    HAMMING = "hamming"
    BARTLETT = "bartlett"
    BLACKMAN = "blackman"
    HANNING = "hanning"


class FourierSpectrum(HasTraits):
    """
    Result of a Fourier  Analysis.
    """
    # Overwrite attribute from superclass
    array_data = NArray(dtype=numpy.complex128)

    source = Attr(
        field_type=time_series.TimeSeries,
        label="Source time-series",
        doc="Links to the time-series on which the FFT is applied.")

    segment_length = Float(
        label="Segment length",
        doc="""The timeseries was segmented into equally sized blocks
            (overlapping if necessary), prior to the application of the FFT.
            The segement length determines the frequency resolution of the
            resulting spectra.""")

    windowing_function = EnumAttr(
        default=WindowingFunctionsEnum.HAMMING,
        required=False,
        label="Windowing function",
        doc="""The windowing function applied to each time segment prior to
            application of the FFT.""")

    amplitude = NArray(label="Amplitude")

    phase = NArray(label="Phase")

    power = NArray(label="Power")

    average_power = NArray(label="Average Power")

    normalised_average_power = NArray(label="Normalised Power", required=False)

    _frequency = None
    _freq_step = None
    _max_freq = None

    def configure(self):
        """ compute dependent fields like amplitude """
        self.compute_amplitude()
        self.compute_phase()
        self.compute_average_power()
        self.compute_normalised_average_power()

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        return {
            "Spectral type": self.__class__.__name__,
            "Source": self.source.title,
            "Segment length": self.segment_length,
            "Windowing function": self.windowing_function,
            "Frequency step": self.freq_step,
            "Maximum frequency": self.max_freq
        }

    @property
    def freq_step(self):
        """ Frequency step size of the complex Fourier spectrum."""
        if self._freq_step is None:
            self._freq_step = 1.0 / self.segment_length
            msg = "%s: Frequency step size is %s"
            self.log.debug(msg % (str(self), str(self._freq_step)))
        return self._freq_step

    @property
    def max_freq(self):
        """ Amplitude of the complex Fourier spectrum."""
        if self._max_freq is None:
            self._max_freq = 0.5 / self.source.sample_period
            msg = "%s: Max frequency is %s"
            self.log.debug(msg % (str(self), str(self._max_freq)))
        return self._max_freq

    @property
    def frequency(self):
        """ Frequencies represented the complex Fourier spectrum."""
        if self._frequency is None:
            self._frequency = numpy.arange(self.freq_step,
                                           self.max_freq + self.freq_step,
                                           self.freq_step)
        return self._frequency

    def compute_amplitude(self):
        """ Amplitude of the complex Fourier spectrum."""
        self.amplitude = numpy.abs(self.array_data)

    def compute_phase(self):
        """ Phase of the Fourier spectrum."""
        self.phase = numpy.angle(self.array_data)

    def compute_power(self):
        """ Power of the complex Fourier spectrum."""
        self.power = numpy.abs(self.array_data) ** 2

    def compute_average_power(self):
        """ Average-power of the complex Fourier spectrum."""
        self.average_power = numpy.mean(numpy.abs(self.array_data) ** 2, axis=-1)

    def compute_normalised_average_power(self):
        """ Normalised-average-power of the complex Fourier spectrum."""
        self.normalised_average_power = (self.average_power /
                                         numpy.sum(self.average_power, axis=0))


class WaveletCoefficients(HasTraits):
    """
    This class bundles all the elements of a Wavelet Analysis into a single
    object, including the input TimeSeries datatype and the output results as
    arrays (FloatArray)
    """
    # Overwrite attribute from superclass
    array_data = NArray(dtype=numpy.complex128)

    source = Attr(field_type=time_series.TimeSeries, label="Source time-series")

    mother = Attr(
        field_type=str,
        label="Mother wavelet",
        default="morlet",
        doc="""A string specifying the type of mother wavelet to use,
            default is 'morlet'.""")  # default to 'morlet'

    sample_period = Float(label="Sample period")
    # sample_rate = basic.Integer(label = "")  inversely related

    frequencies = NArray(
        label="Frequencies",
        doc="A vector that maps scales to frequencies.")

    normalisation = Attr(field_type=str, label="Normalisation type")
    # 'unit energy' | 'gabor'

    q_ratio = Float(label="Q-ratio", default=5.0)

    amplitude = NArray(label="Amplitude")

    phase = NArray(label="Phase")

    power = NArray(label="Power")

    _frequency = None
    _time = None

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        return {
            "Spectral type": self.__class__.__name__,
            "Source": self.source.title,
            "Wavelet type": self.mother,
            "Normalisation": self.normalisation,
            "Q-ratio": self.q_ratio,
            "Sample period": self.sample_period,
            "Number of scales": self.frequencies.shape[0],
            "Minimum frequency": self.frequencies[0],
            "Maximum frequency": self.frequencies[-1]
        }

    @property
    def frequency(self):
        """ Frequencies represented by the wavelet spectrogram."""
        if self._frequency is None:
            self._frequency = numpy.arange(self.frequencies.lo,
                                           self.frequencies.hi,
                                           self.frequencies.step)
        return self._frequency

    def compute_amplitude(self):
        """ Amplitude of the complex Wavelet coefficients."""
        self.amplitude = numpy.abs(self.array_data)

    def compute_phase(self):
        """ Phase of the Wavelet coefficients."""
        self.phase = numpy.angle(self.array_data)

    def compute_power(self):
        """ Power of the complex Wavelet coefficients."""
        self.power = numpy.abs(self.array_data) ** 2


class CoherenceSpectrum(HasTraits):
    """
    Result of a NodeCoherence Analysis.
    """
    # Overwrite attribute from superclass
    array_data = NArray()

    source = Attr(
        field_type=time_series.TimeSeries,
        label="Source time-series",
        doc="""Links to the time-series on which the node_coherence is
            applied.""")

    nfft = Int(
        label="Data-points per block",
        default=256,
        doc="""NOTE: must be a power of 2""")

    frequency = NArray(label="Frequency")

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        return {
            "Spectral type": self.__class__.__name__,
            "Source": self.source.title,
            "Number of frequencies": self.frequency.shape[0],
            "Minimum frequency": self.frequency[0],
            "Maximum frequency": self.frequency[-1],
            "FFT length (time-points)": self.nfft
        }


class ComplexCoherenceSpectrum(HasTraits):
    """
    Result of a NodeComplexCoherence Analysis.
    """

    cross_spectrum = NArray(
        dtype=numpy.complex128,
        label="The cross spectrum",
        doc=""" A complex ndarray that contains the nodes x nodes cross
                spectrum for every frequency frequency and for every segment.""")

    array_data = NArray(
        dtype=numpy.complex128,
        label="Complex Coherence",
        doc="""The complex coherence coefficients calculated from the cross
                spectrum. The imaginary values of this complex ndarray represent the
                imaginary coherence.""")

    source = Attr(
        field_type=time_series.TimeSeries,
        label="Source time-series",
        doc="""Links to the time-series on which the node_coherence is
                applied.""")

    epoch_length = Float(
        label="Epoch length",
        doc="""The timeseries was segmented into equally sized blocks
                (overlapping if necessary), prior to the application of the FFT.
                The segement length determines the frequency resolution of the
                resulting spectra.""")

    segment_length = Float(
        label="Segment length",
        doc="""The timeseries was segmented into equally sized blocks
                (overlapping if necessary), prior to the application of the FFT.
                The segement length determines the frequency resolution of the
                resulting spectra.""")

    windowing_function = Attr(
        field_type=str,
        label="Windowing function",
        doc="""The windowing function applied to each time segment prior to
                application of the FFT.""")

    _frequency = None
    _freq_step = None
    _max_freq = None
    spectrum_types = ["Imaginary", "Real", "Absolute"]

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this datatype.
        """
        return {
            "Spectral type": self.__class__.__name__,
            "Source": self.source.title,
            "Frequency step": self.freq_step,
            "Maximum frequency": self.max_freq,
            "Epoch length": self.epoch_length,
            "Segment length": self.segment_length,
            "Windowing function": self.windowing_function
        }

    @property
    def freq_step(self):
        """ Frequency step size of the Complex Coherence Spectrum."""
        if self._freq_step is None:
            self._freq_step = 1.0 / self.segment_length
            msg = "%s: Frequency step size is %s"
            self.log.debug(msg % (str(self), str(self._freq_step)))
        return self._freq_step

    @property
    def max_freq(self):
        """ Maximum frequency represented in the Complex Coherence Spectrum."""
        if self._max_freq is None:
            self._max_freq = 0.5 / self.source.sample_period
            msg = "%s: Max frequency is %s"
            self.log.debug(msg % (str(self), str(self._max_freq)))
        return self._max_freq

    @property
    def frequency(self):
        """ Frequencies represented in the Complex Coherence Spectrum."""
        if self._frequency is None:
            self._frequency = numpy.arange(self.freq_step,
                                           self.max_freq + self.freq_step,
                                           self.freq_step)
        return self._frequency



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Tests for the CUDA backend.

.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""

import unittest
import numpy as np

from tvb.simulator.backend.cu import CuBackend, pycuda_available
if pycuda_available:  # quickfix
    from tvb.simulator.backend.cu import Out, In, InOut
from tvb.simulator.coupling import Sigmoidal, Linear
from tvb.simulator.models.infinite_theta import MontbrioPazoRoxin

from .backendtestbase import BaseTestSim, BaseTestCoupling, BaseTestDfun


class TestCUSim(BaseTestSim):

    @unittest.skipUnless(pycuda_available, 'requires working PyCUDA')
    def test_mpr(self):
        "Test generated CUDA kernel directly from Simulator instance."
        sim, state, t, y = self._create_sim(inhom_mmpr=True)
        template = '<%include file="cu-sim-ode.cu.mako"/>'
        kernel = CuBackend().build_func(template, dict(sim=sim, pi=np.pi))
        dX = state.copy()
        weights = sim.connectivity.weights.T.copy().astype('f')
        parmat = sim.model.spatial_parameter_matrix.astype('f')
        yh = np.empty((len(t),)+state.shape, 'f')
        kernel(
            In(state), In(weights), Out(yh), In(parmat),
            grid=(1,1), block=(128,1,1))
        self._check_match(y, yh[:,:,0])


class TestCUCoupling(BaseTestCoupling):

    def _test_cfun(self, cfun):
        "Test CUDA cfun template."
        class sim:  # dummy
            model = MontbrioPazoRoxin()
            coupling = cfun
        template = '''
<%include file="cu-coupling.cu.mako"/>
__global__ void kernel(float *state, float *weights, float *cX) {
    coupling(threadIdx.x, ${n_node}, cX, weights, state);
}
'''
        content = dict(n_node=128, sim=sim)
        kernel = CuBackend().build_func(template, content)
        state = np.random.rand(2, content['n_node']).astype('f')
        weights = np.random.randn(state.shape[1], state.shape[1]).astype('f')
        cX = np.empty_like(state)
        kernel(In(state), In(weights), Out(cX), 
            grid=(1,1), block=(content['n_node'],1,1))
        expected = self._eval_cfun_no_delay(sim.coupling, weights, state)
        np.testing.assert_allclose(cX, expected, 1e-5, 1e-6)

    @unittest.skipUnless(pycuda_available, 'requires working PyCUDA')
    def test_linear(self):
        self._test_cfun(Linear())

    @unittest.skipUnless(pycuda_available, 'requires working PyCUDA')
    def test_sigmoidal(self):
        self._test_cfun(Sigmoidal())


class TestCUDfun(BaseTestDfun):

    def _test_model(self, model_):
        "Test CUDA model dfuns."
        class sim:  # dummy
            model = model_
        template = '''

#define M_PI_F 3.14159265358979f

<%include file="cu-dfuns.cu.mako"/>
__global__ void kernel(float *dX, float *state, float *cX, float *parmat) {
    dfuns(threadIdx.x, ${n_node}, dX, state, cX, parmat);
}
'''
        content = dict(n_node=128, sim=sim)
        kernel = CuBackend().build_func(template, content, print_source=True)
        dX, state, cX = np.random.rand(3, 2, content['n_node']).astype('f')
        parmat = sim.model.spatial_parameter_matrix.astype('f')
        if parmat.size == 0:
            parmat = np.zeros((1,),'f') # dummy
        kernel(Out(dX), In(state), In(cX), In(parmat), 
            grid=(1,1), block=(content['n_node'],1,1))
        expected = sim.model.dfun(state, cX)
        np.testing.assert_allclose(dX, expected, 1e-3, 1e-5)

    @unittest.skipUnless(pycuda_available, 'requires working PyCUDA')
    def test_mpr_symmetric(self):
        self._test_model(self._prep_model())

    @unittest.skipUnless(pycuda_available, 'requires working PyCUDA')
    def test_mpr_spatial1(self):
        "Test MPR w/ 1 spatial parameter."
        self._test_model(self._prep_model(1))

    @unittest.skipUnless(pycuda_available, 'requires working PyCUDA')
    def test_mpr_spatial2(self):
        "Test MPR w/ 2 spatial parameters."
        self._test_model(self._prep_model(2))



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Calculate a wavelet transform on a TimeSeries datatype and return a
WaveletSpectrum datatype.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>
.. moduleauthor:: Andreas Spiegler <anspiegler@googlemail.com>
.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>
.. moduleauthor:: Paula Sanz Leon <Paula@tvb.invalid>

"""

import numpy
import scipy.signal as signal
import tvb.datatypes.spectral as spectral
from tvb.basic.logger.builder import get_logger
from tvb.basic.neotraits.api import HasTraits, Attr, Range, Float, narray_describe
from tvb.simulator.backend.ref import ReferenceBackend

SUPPORTED_WAVELET_FUNCTIONS = ("morlet",)

log = get_logger(__name__)

"""
A module for calculating the wavelet transform of a TimeSeries object of TVB
and returning a WaveletSpectrum object. The sampling period and frequency
range of the result can be specified. The mother wavelet can also be
specified... (So far, only Morlet.)

References:
    .. [TBetal_1996] C. Tallon-Baudry et al, *Stimulus Specificity of
        Phase-Locked and Non-Phase-Locked 40 Hz Visual Responses in Human.*,
        J Neurosci 16(13):4240-4249, 1996.

    .. [Mallat_1999] S. Mallat, *A wavelet tour of signal processing.*,
        book, Academic Press, 1999.
"""


def compute_continuous_wavelet_transform(time_series, frequencies, sample_period, q_ratio, normalisation, mother):
    """
    # type: (TimeSeries, Range, float, float, str, str)  -> WaveletCoefficients
    Calculate the continuous wavelet transform of time_series.

    Parameters
    __________

    time_series : TimeSeries
    The timeseries to which the wavelet is to be applied.

    frequencies : Range
    The frequency resolution and range returned. Requested frequencies
    are expected to be in kHz.

    sample_period : float
    The sampling period in ms of the computed wavelet spectrum.

    q_ratio : float
    NFC. Must be greater than 5. Ratios of the center frequencies to bandwidths.

    normalisation : str
    The type of normalisation for the resulting wavet spectrum. Default is 'energy', options are: 'energy'; 'gabor'.

    mother : str
    The mother wavelet function used in the transform.
    """
    ts_shape = time_series.data.shape

    if frequencies.step == 0:
        log.warning("Frequency step can't be 0! Trying default step, 2e-3.")
        frequencies.step = 0.002

    freqs = numpy.arange(frequencies.lo, frequencies.hi, frequencies.step)

    if (freqs.size == 0) or any(freqs <= 0.0):
        # TODO: Maybe should limit number of freqs... ~100 is probably a reasonable upper bound.
        log.warning("Invalid frequency range! Falling back to default.")
        log.debug("freqs")
        log.debug(narray_describe(freqs))
        frequencies = Range(lo=0.008, hi=0.060, step=0.002)
        freqs = numpy.arange(frequencies.lo, frequencies.hi, frequencies.step)

    log.debug("freqs")
    log.debug(narray_describe(freqs))

    # We need this to be kHz (see TVB-2946)
    sample_rate = time_series.sample_rate / 1000

    # Duke: code below is as given by Andreas Spiegler, I've just wrapped
    # some of the original argument names
    nf = len(freqs)
    temporal_step = max((1, ReferenceBackend.iround(sample_period / time_series.sample_period_ms)))
    nt = int(numpy.ceil(ts_shape[0] / temporal_step))

    if not isinstance(q_ratio, numpy.ndarray):
        new_q_ratio = q_ratio * numpy.ones((1, nf))

    if numpy.nanmin(new_q_ratio) < 5:
        msg = "q_ratio must be not lower than 5 !"
        log.error(msg)
        raise Exception(msg)

    if numpy.nanmax(freqs) > sample_rate / 2.0:
        msg = "Sampling rate is too low for the requested frequency range !"
        log.error(msg)
        raise Exception(msg)

    # TODO: This isn't used, but min frequency seems like it should be important... Check with A.S.
    #  fmin = 3.0 * numpy.nanmin(q_ratio) * sample_rate / numpy.pi / nt
    sigma_f = freqs / new_q_ratio
    sigma_t = 1.0 / (2.0 * numpy.pi * sigma_f)

    if normalisation == 'energy':
        Amp = 1.0 / numpy.sqrt(sample_rate * numpy.sqrt(numpy.pi) * sigma_t)
    elif normalisation == 'gabor':
        Amp = numpy.sqrt(2.0 / numpy.pi) / sample_rate / sigma_t

    coef_shape = (nf, nt, ts_shape[1], ts_shape[2], ts_shape[3])

    coef = numpy.zeros(coef_shape, dtype=numpy.complex128)
    log.debug("coef")
    log.debug(narray_describe(coef))

    scales = numpy.arange(0, nf, 1)
    for i in scales:
        f0 = freqs[i]
        SDt = sigma_t[(0, i)]
        A = Amp[(0, i)]
        x = numpy.arange(0, 4.0 * SDt * sample_rate, 1) / sample_rate
        wvlt = A * numpy.exp(-x ** 2 / (2.0 * SDt ** 2)) * numpy.exp(2j * numpy.pi * f0 * x)
        wvlt = numpy.hstack((numpy.conjugate(wvlt[-1:0:-1]), wvlt))
        # util.self.log_debug_array(self.log, wvlt, "wvlt")

        for var in range(ts_shape[1]):
            for node in range(ts_shape[2]):
                for mode in range(ts_shape[3]):
                    data = time_series.data[:, var, node, mode]
                    wt = signal.convolve(data, wvlt, 'same')
                    # util.self.log_debug_array(self.log, wt, "wt")
                    res = wt[0::temporal_step]
                    # NOTE: this is a horrible horrible quick hack (alas, a solution) to avoid broadcasting errors
                    # when using dt and sample periods which are not powers of 2.
                    coef[i, :, var, node, mode] = res if len(res) == nt else res[:coef.shape[1]]

    log.debug("coef")
    log.debug(narray_describe(coef))

    spectra = spectral.WaveletCoefficients(
        source=time_series,
        mother=mother,
        sample_period=sample_period,
        frequencies=frequencies.to_array(),
        normalisation=normalisation,
        q_ratio=q_ratio,
        array_data=coef)

    return spectra



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Functions that inform a user about the state of a traited class or object.

Some of these functions are here so that they won't clutter the core trait implementation.
"""

import uuid
import numpy
import typing

try:
    from docutils.core import publish_parts
except ImportError:
    def publish_parts(param1, _param2):
        return param1


def auto_docstring(cls):
    """ generate a docstring for the new class in which the Attrs are documented """
    header = 'Traited class [{}.{}]'.format(cls.__module__, cls.__name__)

    doc = [
        header,
        len(header) * '^',
        '',
    ]
    if cls.__doc__ is not None:
        doc.extend([cls.__doc__, ''])

    doc.extend([
        'Attributes declared',
        '"""""""""""""""""""',
        ''
    ])
    # a rst definition list for all attributes
    for attr_name in cls.declarative_attrs:
        attr = getattr(cls, attr_name)
        # the standard repr of the attribute
        doc.append('{} : {}'.format(attr_name, str(attr)))
        # and now the doc property
        for line in attr.doc.splitlines():
            doc.append('    ' + line.lstrip())
        doc.append('')

    if cls.declarative_props:
        doc.extend([
            '',
            'Properties declared',
            '"""""""""""""""""""',
            ''
        ])

    for prop_name in cls.declarative_props:
        prop = getattr(cls, prop_name)
        # the standard repr
        doc.append('  {} : {}'.format(prop_name, str(prop)))
        # now fish the docstrings
        for line in prop.attr.doc.splitlines():
            doc.append('    ' + line.lstrip())
        if prop.fget.__doc__ is not None:
            for line in prop.fget.__doc__.splitlines():
                doc.append('    ' + line.lstrip())

    doc = '\n'.join(doc)

    return doc


def narray_summary_info(ar, ar_name='', condensed=False):
    # type: (numpy.ndarray, str, bool) -> typing.Dict[str, str]
    """
    A 2 column table represented as a dict of str->str
    """

    key_none = 'is None'
    key_empty = 'is empty'
    key_min_max = '[min, median, max]'
    key_nan = 'has NaN'
    key_shape = 'shape'
    key_type = 'dtype'

    if ar is None:
        return {f'{ar_name} {key_none}': 'True'} if ar_name else {key_none: 'True'}

    if ar.size == 0:
        return {f'{ar_name} {key_empty}': 'True'} if ar_name else {key_empty: 'True'}

    ret = {key_shape: str(ar.shape),
           key_type: str(ar.dtype)}

    if ar.dtype.kind in 'iufc':
        has_nan = numpy.isnan(ar).any()
        if has_nan:
            ret[key_nan] = 'True'
        ret[key_min_max] = '[{:g}, {:g}, {:g}]'.format(ar.min(), numpy.median(ar), ar.max())

    if condensed:
        condensed_desc = ""
        if ar.shape == (1,):
            condensed_desc += str(ar.item())
        else:
            for key in [key_min_max, key_type, key_shape, key_empty, key_nan]:
                if key in ret:
                    condensed_desc += f' {key} = {ret[key]}'

        return {ar_name: condensed_desc} if ar_name else {'array': condensed_desc}

    if ar_name:
        return {ar_name + ' ' + k: v for k, v in ret.items()}
    else:
        return ret


def narray_describe(ar):
    # type: (numpy.ndarray) -> str
    summary = narray_summary_info(ar)
    ret = []
    for k in sorted(summary):
        ret.append('{:<12}{}'.format(k, summary[k]))
    return '\n'.join(ret)


# these are here and not on HasTraits just so that that class is not
# complicated by irrelevant string formatting


def trait_object_str(self):
    cls = type(self)
    summary = self.summary_info()
    result = ['{} ('.format(cls.__name__)]
    maxlenk = max(len(k) for k in summary)

    for k in sorted(summary):
        result.append('  {:.<{}} {}'.format(k + ' ', maxlenk, summary[k]))
    result.append(')')
    return '\n'.join(result)


def trait_object_repr_html(self):
    cls = type(self)

    subtitle = None
    if hasattr(self, "dfun"):
        subtitle = self.dfun.__doc__
    elif hasattr(cls, "__doc_old__"):
        subtitle = cls.__doc_old__

    result = [
        '<table>',
        '<thead><h3>{}</h3></thead>'.format(cls.__name__),
        '<tbody>',
        '<tr><td colspan="2"><p>{}</p></td></tr>'.format(prepare_html(subtitle)) if subtitle is not None else "",
        '<tr><th></th><th style="text-align:left;width:80%">value</th></tr>',
    ]

    summary = self.summary_info()

    for k in sorted(summary):
        row_fmt = '<tr><td>{}</td><td style="text-align:left;"><pre>{}</pre></td></tr>'
        result.append(row_fmt.format(k, summary[k]))

    result += ['</tbody></table>']

    return '\n'.join(result)


def convert_rst_to_html(doc):
    """
    Convert from rst to html that can be rendered by Mathjax
    """
    kwargs = {
        'writer_name': 'html',
        'settings_overrides': {
            '_disable_config': True,
            'report_level': 5,
            'math_output': "MathJax /dummy.js",
        },
    }

    return publish_parts(doc, **kwargs)['html_body']


def prepare_html(doc):
    # type: (str) -> str
    """
    Create html (that can be further enhanced by MathJax) from the description received as parameter
    """
    try:
        html_id = uuid.uuid1()

        html = convert_rst_to_html(doc)

        html = html.replace('div class="document"', f'div class="document" id="{html_id}"', 1)
        html += fr'<script>MathJax.Hub.Queue(["Typeset", MathJax.Hub, "{html_id}"]);</script>'

    except Exception:
        html = str(doc)
    return html



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
.. moduleauthor:: Dionysios Perdikis <dionperd@gmail.com>
.. moduleauthor:: Gabriel Florea <gabriel.florea@codemart.ro>
"""

import numpy as np
import os
from scipy.interpolate import interp1d, griddata
from scipy.signal import welch, periodogram, spectrogram
from six import string_types

from tvb.basic.logger.builder import get_logger
from tvb.datatypes.connectivity import Connectivity
from tvb.simulator.coupling import Linear
from tvb.simulator.integrators import HeunStochastic
from tvb.simulator.models.oscillator import Generic2dOscillator
from tvb.simulator.monitors import TemporalAverage
from tvb.simulator.noise import Additive
from tvb.simulator.plot.config import CONFIGURED
from tvb.simulator.simulator import Simulator

LOG = get_logger(__name__)


# Analyzers utils_pack:

def spectral_analysis(x, fs, freq=None, method="periodogram", output="spectrum", nfft=None, window='hanning',
                      nperseg=256, detrend='constant', noverlap=None, f_low=10.0, log_scale=False):
    if freq is None:
        freq = np.linspace(f_low, nperseg, nperseg - f_low - 1)
        df = freq[1] - freq[0]
    psd = []
    for iS in range(x.shape[1]):
        if method is welch:
            f, temp_psd = welch(x[:, iS],
                                fs=fs,  # sample rate
                                nfft=nfft,
                                window=window,  # apply a Hanning window before taking the DFT
                                nperseg=nperseg,  # compute periodograms of 256-long segments of x
                                detrend=detrend,
                                scaling="spectrum",
                                noverlap=noverlap,
                                return_onesided=True,
                                axis=0)
        else:
            f, temp_psd = periodogram(x[:, iS],
                                      fs=fs,  # sample rate
                                      nfft=nfft,
                                      window=window,  # apply a Hanning window before taking the DFT
                                      detrend=detrend,
                                      scaling="spectrum",
                                      return_onesided=True,
                                      axis=0)
        f = interp1d(f, temp_psd)
        temp_psd = f(freq)
        if output == "density":
            temp_psd /= (np.sum(temp_psd) * df)
        psd.append(temp_psd)
    # Stack them to a ndarray
    psd = np.stack(psd, axis=1)
    if output == "energy":
        return np.sum(psd, axis=0)
    else:
        if log_scale:
            psd = np.log(psd)
        return psd, freq


def time_spectral_analysis(x, fs, freq=None, mode="psd", nfft=None, window='hanning', nperseg=256, detrend='constant',
                           noverlap=None, f_low=10.0, calculate_psd=True, log_scale=False):
    # TODO: add a Continuous Wavelet Transform implementation
    if freq is None:
        freq = np.linspace(f_low, nperseg, nperseg - f_low - 1)
    stf = []
    for iS in range(x.shape[1]):
        f, t, temp_s = spectrogram(x[:, iS], fs=fs, nperseg=nperseg, nfft=nfft, window=window, mode=mode,
                                   noverlap=noverlap, detrend=detrend, return_onesided=True, scaling='spectrum', axis=0)
        t_mesh, f_mesh = np.meshgrid(t, f, indexing="ij")
        temp_s = griddata((t_mesh.flatten(), f_mesh.flatten()), temp_s.T.flatten(),
                          tuple(np.meshgrid(t, freq, indexing="ij")), method='linear')
        stf.append(temp_s)
    # Stack them to a ndarray
    stf = np.stack(stf, axis=2)
    if log_scale:
        stf = np.log(stf)
    if calculate_psd:
        psd, _ = spectral_analysis(x, fs, freq=freq, method="periodogram", output="spectrum", nfft=nfft, window=window,
                                   nperseg=nperseg, detrend=detrend, noverlap=noverlap, log_scale=log_scale)
        return stf, t, freq, psd
    else:
        return stf, t, freq


# Data structures utils

def isequal_string(a, b, case_sensitive=False):
    if case_sensitive:
        return a == b
    else:
        try:
            return a.lower() == b.lower()
        except AttributeError:
            LOG.warning("Case sensitive comparison!")
            return a == b


def ensure_list(arg):
    if not (isinstance(arg, list)):
        try:  # if iterable
            if isinstance(arg, (str, dict)):
                arg = [arg]
            elif hasattr(arg, "__iter__"):
                arg = list(arg)
            else:  # if not iterable
                arg = [arg]
        except:  # if not iterable
            arg = [arg]
    return arg


def ensure_string(arg):
    if not (isinstance(arg, string_types)):
        if arg is None:
            return ""
        else:
            return ensure_list(arg)[0]
    else:
        return arg


def generate_region_labels(n_regions, labels=[], str=". ", numbering=True, numbers=[]):
    if len(numbers) != n_regions:
        numbers = list(range(n_regions))
    if len(labels) == n_regions:
        if numbering:
            return np.array([str.join(["%d", "%s"]) % tuple(l) for l in zip(numbers, labels)])
        else:
            return np.array(labels)
    else:
        return np.array(["%d" % l for l in numbers])


# Computational utils

def compute_in_degree(weights):
    return np.expand_dims(np.sum(weights, axis=1), 1).T


def normalize_weights(weights, percentile=CONFIGURED.WEIGHTS_NORM_PERCENT, remove_diagonal=True, ceil=1.0):
    # Create the normalized connectivity weights:
    if len(weights) > 0:
        normalized_w = np.array(weights)
        if remove_diagonal:
            # Remove diagonal elements
            n_regions = normalized_w.shape[0]
            normalized_w *= (1.0 - np.eye(n_regions))
        # Normalize with the 95th percentile
        normalized_w = np.array(normalized_w / np.percentile(normalized_w, percentile))
        if ceil:
            if ceil is True:
                ceil = 1.0
            normalized_w[normalized_w > ceil] = ceil
        return normalized_w
    else:
        return np.array([])


def raise_value_error(msg, logger=None):
    if logger is not None:
        logger.error("\n\nValueError: " + msg + "\n")
    raise ValueError(msg)


def rotate_n_list_elements(lst, n):
    lst = ensure_list(lst)
    n_lst = len(lst)
    if n_lst != n and n_lst != 0:
        if n_lst == 1:
            lst *= n
        elif n_lst > n:
            lst = lst[:n]
        else:
            old_lst = list(lst)
            while n_lst < n:
                lst += old_lst[0]
                old_lst = old_lst[1:] + old_lst[:1]
    return lst


# if the demo data are not generated, this function will.
def generate_region_demo_data(file_path=os.path.join(os.getcwd(), "demo_data_region_16s_2048Hz.npy")):
    """
    Generate 16 seconds of 2048Hz data at the region level, stochastic integration.

    ``Run time``: approximately 4 minutes (workstation circa 2010)

    ``Memory requirement``: < 1GB
    ``Storage requirement``: ~ 19MB

    .. moduleauthor:: Stuart A. Knock <stuart.knock@gmail.com>

    """

    ##----------------------------------------------------------------------------##
    ##-                      Perform the simulation                              -##
    ##----------------------------------------------------------------------------##

    LOG.info("Configuring...")

    # Initialise a Model, Coupling, and Connectivity.
    pars = {'a': np.array([1.05]),
            'b': np.array([-1]),
            'c': np.array([0.0]),
            'd': np.array([0.1]),
            'e': np.array([0.0]),
            'f': np.array([1 / 3.]),
            'g': np.array([1.0]),
            'alpha': np.array([1.0]),
            'beta': np.array([0.2]),
            'tau': np.array([1.25]),
            'gamma': np.array([-1.0])}

    oscillator = Generic2dOscillator(**pars)

    white_matter = Connectivity.from_file()
    white_matter.speed = np.array([4.0])
    white_matter_coupling = Linear(a=np.array([0.033]))

    # Initialise an Integrator
    hiss = Additive(nsig=np.array([2 ** -10, ]))
    heunint = HeunStochastic(dt=0.06103515625, noise=hiss)

    # Initialise a Monitor with period in physical time
    what_to_watch = TemporalAverage(period=0.48828125)  # 2048Hz => period=1000.0/2048.0

    # Initialise a Simulator -- Model, Connectivity, Integrator, and Monitors.
    sim = Simulator(model=oscillator, connectivity=white_matter,
                    coupling=white_matter_coupling,
                    integrator=heunint, monitors=[what_to_watch])

    sim.configure()

    # Perform the simulation
    tavg_data = []
    tavg_time = []
    LOG.info("Starting simulation...")
    for tavg in sim(simulation_length=16000):
        if tavg is not None:
            tavg_time.append(tavg[0][0])  # TODO:The first [0] is a hack for single monitor
            tavg_data.append(tavg[0][1])  # TODO:The first [0] is a hack for single monitor

    LOG.info("Finished simulation.")

    ##----------------------------------------------------------------------------##
    ##-                     Save the data to a file                              -##
    ##----------------------------------------------------------------------------##

    # Make the list a numpy.array.
    LOG.info("Converting result to array...")
    TAVG = np.array(tavg_data)

    # Save it
    LOG.info("Saving array to %s..." % file_path)
    np.save(file_path, TAVG)

    LOG.info("Done.")



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""

The Pattern datatypes.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

import numpy

from tvb.datatypes import surfaces, volumes, connectivity, equations
from tvb.basic.neotraits.api import HasTraits, NArray, Attr


class SpatialPattern(HasTraits):
    """
    Equation for space variation.
    """

    spatial =  Attr(field_type=equations.FiniteSupportEquation, label="Spatial Equation")

    space = None
    _spatial_pattern = None

    def summary_info(self):
        """
        Gather scientifically interesting summary information from an instance of this DataType.
        """
        return {"Type": self.__class__.__name__,
                "Spatial equation": self.spatial.__class__.__name__,
                "Spatial parameters": self.spatial.parameters}

    @property
    def spatial_pattern(self):
        """
        Return a discrete representation of the spatial pattern.
        """
        return self._spatial_pattern

    def configure_space(self, distance):
        """
        Stores the distance vector as an attribute of the spatiotemporal pattern
        and uses it to generate the spatial pattern vector.

        Depending on equations used and interpretation distance can be an actual
        physical distance, on a surface,  geodesic distance (along the surface)
        away for some focal point, or a per node weighting...
        """
        # Set the discrete representation of space.
        self.space = distance
        # Generate a discrete representation of the spatial pattern.
        # The argument x represents a distance, or effective distance, for each node in the space.
        self._spatial_pattern = numpy.sum(self.spatial.evaluate(self.space), axis=1)[:, numpy.newaxis]


class SpatioTemporalPattern(SpatialPattern):
    """
    Combine space and time equations.
    """

    temporal = Attr(field_type=equations.TemporalApplicableEquation, label="Temporal Equation")
    # space must be shape (x, 1); time must be shape (1, t)
    time = None
    _temporal_pattern = None

    def summary_info(self):
        """ Extend the base class's summary dictionary. """
        summary = super(SpatioTemporalPattern, self).summary_info()
        summary["Temporal equation"] = self.temporal.__class__.__name__
        summary["Temporal parameters"] = self.temporal.parameters
        return summary

    @property
    def temporal_pattern(self):
        """
        Return a discrete representation of the temporal pattern.
        """
        return self._temporal_pattern

    def configure_time(self, time):
        """
        Stores the time vector, physical units (ms), as an attribute of the
        spatio-temporal pattern and uses it to generate the temporal pattern
        vector.
        """
        self.time = time
        # Generate a discrete representation of the temporal pattern.
        self._temporal_pattern = numpy.reshape(self.temporal.evaluate(self.time), (1, -1))

    def __call__(self, temporal_indices=None, spatial_indices=None):
        """
        The temporal pattern vector, set by the configure_time method, is
        combined with the spatial pattern vector, set by the configure_space
        method, to form a spatiotemporal pattern.

        Called with a single time index as an argument, the spatial pattern at
        that point in time is returned. This is the standard usage within a
        simulation where the current simulation time point is retrieved.

        Called without any arguments, by default a big array representing the
        entire spatio-temporal pattern is returned. While this may be useful for
        visualisation, say of region level spatio-temporal patterns, care should
        be taken as when surfaces are considered the returned array can be
        potentially quite large.
        """
        pattern = None
        if temporal_indices is not None and spatial_indices is None:
            pattern = self._spatial_pattern * self._temporal_pattern[0, temporal_indices]
        elif temporal_indices is None and spatial_indices is None:
            pattern = self._spatial_pattern * self._temporal_pattern
        elif temporal_indices is not None and spatial_indices is not None:
            pattern = self._spatial_pattern[spatial_indices, 0] * self._temporal_pattern[0, temporal_indices]
        elif temporal_indices is None and spatial_indices is not None:
            pattern = self._spatial_pattern[spatial_indices, 0] * self._temporal_pattern
        else:
            self.log.error("%s: Well, that shouldn't be possible..." % repr(self))
        return pattern


class StimuliRegion(SpatioTemporalPattern):
    """
    A class that bundles the temporal profile of the stimulus, together with the
    list of scaling weights of the regions where it will applied.
    """
    connectivity = Attr(field_type=connectivity.Connectivity, label="Connectivity")

    spatial = Attr(field_type=equations.DiscreteEquation,
                   label="Spatial Equation",
                   default=equations.DiscreteEquation())  # fixed_type=True, order=-1)

    weight = NArray(label="scaling")  # , locked=True, order=4)

    @staticmethod
    def get_default_weights(number_of_regions):
        """
        Returns a list with a number of elements
        equal to the given number of regions.
        """
        return numpy.array([0.0] * number_of_regions)

    @property
    def weight_array(self):
        """
        Wrap weight List into a Numpy array, as it is requested by the simulator.
        """
        return numpy.array(self.weight)[:, numpy.newaxis]

    def configure_space(self, region_mapping=None):
        """
        Do necessary preparations in order to use this stimulus.
        NOTE: this was previously done in simulator configure_stimuli() method.
        It no needs to be used in stimulus viewer also.
        """
        if region_mapping is not None:
            # TODO: smooth at surface region boundaries
            distance = self.weight_array[region_mapping, :]
        else:
            distance = self.weight_array
        super(StimuliRegion, self).configure_space(distance)


class StimuliSurface(SpatioTemporalPattern):
    """
    A spatio-temporal pattern defined in a Surface DataType.
    It includes the list of focal points.
    """

    surface = Attr(field_type=surfaces.CorticalSurface, label="Surface")

    focal_points_triangles = NArray(dtype=int, label="Focal points triangles")  # , locked=True, order=4)

    @property
    def focal_points_surface(self):
        focal_points = []

        if self.surface is None or self.surface.triangles is None:
            self.log.warning('Focal points list will be empty. Load the surface triangles before accessing this property!')
            return numpy.array(focal_points)

        for triangle_index in self.focal_points_triangles:
            focal_points.append(int(self.surface.triangles[triangle_index][0]))
        return numpy.array(focal_points)

    def configure_space(self, region_mapping=None):
        """
        Do necessary preparations in order to use this stimulus.
        NOTE: this was previously done in simulator configure_stimuli() method.
        It no needs to be used in stimulus viewer also.
        """
        dis_shp = (self.surface.number_of_vertices, numpy.size(self.focal_points_surface))
        # TODO: When this was in Simulator it was number of nodes, using surface vertices
        # breaks surface simulations which include non-cortical regions.

        distance = numpy.zeros(dis_shp)
        k = -1
        for focal_point in self.focal_points_surface:
            k += 1
            foci = numpy.array([focal_point], dtype=numpy.int32)
            distance[:, k] = self.surface.geodesic_distance(foci)
        super(StimuliSurface, self).configure_space(distance)


class SpatialPatternVolume(SpatialPattern):
    """ A spatio-temporal pattern defined in a volume. """

    volume = Attr(volumes.Volume, label="Volume")

    focal_points_volume = NArray(dtype=int, label="Focal points")



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
   A family of mean field models of infinite populations of all-to-all coupled
   quadratic integrate and fire neurons (theta-neurons).

.. moduleauthor:: Jan Fousek <jan.fousek@univ-amu.fr>, Giovanni Rabuffo <giovanni.rabuffo@univ-amu.fr>
"""

from tvb.simulator.models.base import Model
from tvb.basic.neotraits.api import NArray, List, Range, Final

import numpy


class MontbrioPazoRoxin(Model):
    r"""
    2D model describing the Ott-Antonsen reduction of infinite all-to-all
    coupled QIF neurons (Theta-neurons) as in [Montbrio_Pazo_Roxin_2015]_.

    The two state variables :math:`r` and :math:`V` represent the average
    firing rate and the average membrane potential of our QIF neurons.

    The equations of the infinite QIF 2D population model read

    .. math::
            \dot{r} &= 1/\tau (\Delta/(\pi \tau) + 2 V r)\\
            \dot{V} &= 1/\tau (V^2 - \tau^2 \pi^2 r^2 + \eta + J \tau r + I)
    
    Input from the network enters in the :math:`V` variable as 
    :math:`1/\tau(c_r C_r + c_v C_V)` where C is the incomming coupling. In 
    other words, depending on the parameters :math:`c_r`, :math:`c_v` we couple
    the neural masses via the firing rate and/or the membrane potential.
    
    .. [Montbrio_Pazo_Roxin_2015] Montbrió, E., Pazó, D., & Roxin, A. (2015). Macroscopic description for networks of spiking neurons. *Physical Review X*, 5(2), 021028.
    """

    # Define traited attributes for this model, these represent possible kwargs.

    tau = NArray(
        label=r":math:`\tau`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.001, hi=15.0, step=0.01),
        doc="""Characteristic time""",
    )

    I = NArray(
        label=":math:`I_{ext}`",
        default=numpy.array([0.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.01),
        doc="""External Current""",
    )

    Delta = NArray(
        label=r":math:`\Delta`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.0, hi=10.0, step=0.01),
        doc="""Mean heterogeneous noise""",
    )

    J = NArray(
        label=":math:`J`",
        default=numpy.array([15.0]),
        domain=Range(lo=-25.0, hi=25.0, step=0.0001),
        doc="""Mean Synaptic weight.""",
    )

    eta = NArray(
        label=r":math:`\eta`",
        default=numpy.array([-5.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.0001),
        doc="""Constant parameter to scale the rate of feedback from the
            firing rate variable to itself""",
    )

    Gamma = NArray(
        label=r":math:`\Gamma`",
        default=numpy.array([0.]),
        domain=Range(lo=0., hi=10.0, step=0.01),
        doc="""Half-width of synaptic weight distribution""",
    )

    cr = NArray(
        label=":math:`cr`",
        default=numpy.array([1.]),
        domain=Range(lo=0., hi=1, step=0.1),
        doc="""It is the weight on Coupling through variable r.""",
    )

    cv = NArray(
        label=":math:`cv`",
        default=numpy.array([0.]),
        domain=Range(lo=0., hi=1, step=0.1),
        doc="""It is the weight on Coupling through variable V.""",
    )

    # Informational attribute, used for phase-plane and initial()
    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={"r": numpy.array([0., 2.0]),
                 "V": numpy.array([-2.0, 1.5])},
        doc="""Expected ranges of the state variables for initial condition generation and phase plane setup.""",
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={
            "r": numpy.array([0.0, numpy.inf])
        },
    )

    # TODO should match cvars below..
    coupling_terms = Final(
        label="Coupling terms",
        # how to unpack coupling array
        default=["Coupling_Term_r", "Coupling_Term_V"]
    )

    state_variable_dfuns = Final(
        label="Drift functions",
        default={
            "r": "1/tau * ( Delta / (pi * tau) + 2 * V * r)",
            "V": "1/tau * ( V*V - pi*pi*tau*tau*r*r + eta + J * tau * r + I + cr * Coupling_Term_r + cv * Coupling_Term_V)"
        }
    )

    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=("r", "V"),
        default=("r", "V"),
        doc="The quantities of interest for monitoring for the Infinite QIF 2D oscillator.",
    )

    parameter_names = List(
        of=str,
        label="List of parameters for this model",
        default='tau Delta eta J I cr cv'.split())

    state_variables = ('r', 'V')
    _nvar = 2
    # Cvar is the coupling variable. 
    cvar = numpy.array([0, 1], dtype=numpy.int32)
    # Stvar is the variable where stimulus is applied.
    stvar = numpy.array([1], dtype=numpy.int32)

    def dfun(self, state_variables, coupling, local_coupling=0.0):
        r"""
            2D model describing the Ott-Antonsen reduction of infinite all-to-all
            coupled QIF neurons (Theta-neurons) as in [Montbrio_Pazo_Roxin_2015]_.

            The two state variables :math:`r` and :math:`V` represent the average
            firing rate and the average membrane potential of our QIF neurons.

            The equations of the infinite QIF 2D population model read

            .. math::
                    \dot{r} &= 1/\tau (\Delta/(\pi \tau) + 2 V r)\\
                    \dot{V} &= 1/\tau (V^2 - \tau^2 \pi^2 r^2 + \eta + J \tau r + I)
        """

        r, V = state_variables

        # [State_variables, nodes]
        I = self.I
        Delta = self.Delta
        Gamma = self.Gamma
        eta = self.eta
        tau = self.tau
        J = self.J
        cr = self.cr
        cv = self.cv

        Coupling_Term_r = coupling[0, :]  # This zero refers to the first element of cvar (r in this case)
        Coupling_Term_V = coupling[1, :]  # This zero refers to the second element of cvar (V in this case)

        derivative = numpy.empty_like(state_variables)

        derivative[0] = 1 / tau * (Delta / (numpy.pi * tau) + 2 * V * r)
        derivative[1] = 1 / tau * (
                    V ** 2 - numpy.pi ** 2 * tau ** 2 * r ** 2 + eta + J * tau * r + I + cr * Coupling_Term_r + cv * Coupling_Term_V)

        return derivative


class CoombesByrne(Model):
    r"""
    4D model describing the Ott-Antonsen reduction of infinite all-to-all
    coupled QIF neurons (Theta-neurons) as in [Coombes_Byrne_2019]_.
    
    Note: the original equations describe the dynamics of the Kuramoto parameter 
    :math:`Z`. Using the conformal transformation 
    :math:`Z=(1-W^\star)/(1+W^\star)` and :math:`W= \pi r + i V`, 
    we express the system dynamics in terms of two state variables :math:`r` 
    and :math:`V` representing the average firing rate and the average membrane 
    potential of our QIF neurons. The conductance variable and its derivative 
    are :math:`g` and :math:`q`.

    The equations of the model read
    
    .. math::
            \dot{r} &= \Delta/\pi + 2 V r - g r^2 \\
            \dot{V} &= V^2 - \pi^2 r^2 + \eta + (v_{syn} - V) g \\
            \dot{g} &= \alpha q  \\
            \dot{q} &= \alpha (\kappa \pi r - g - 2 q)
            
    .. [Coombes_Byrne_2019] Coombes, S., & Byrne, Á. (2019). Next generation neural mass models. In *Nonlinear Dynamics in Computational Neuroscience* (pp. 1-16). Springer, Cham.
            
    """

    # Define traited attributes for this model, these represent possible kwargs.

    Delta = NArray(
        label=r":math:`\Delta`",
        default=numpy.array([0.5]),
        domain=Range(lo=0.0, hi=10.0, step=0.01),
        doc="""Half-width of heterogeneous noise distribution""",
    )

    alpha = NArray(
        label=r":math:`\alpha`",
        default=numpy.array([0.95]),
        domain=Range(lo=0.0, hi=10.0, step=0.01),
        doc="""Parameter of the alpha-function""",
    )

    v_syn = NArray(
        label=":math:`v_{syn}`",
        default=numpy.array([-10.0]),
        domain=Range(lo=-20.0, hi=0.0, step=0.01),
        doc="""QIF membrane reversal potential""",
    )

    k = NArray(
        label=":math:`k`",
        default=numpy.array([1.]),
        domain=Range(lo=0., hi=5.0, step=0.01),
        doc="""Local coupling strength""",
    )

    eta = NArray(
        label=r":math:`\eta`",
        default=numpy.array([20.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.0001),
        doc="""Constant parameter to scale the rate of feedback from the
            firing rate variable to itself""",
    )

    # Informational attribute, used for phase-plane and initial()
    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={
            "r": numpy.array([0., 6.]),
            "V": numpy.array([-10., 10.]),
            "g": numpy.array([1., 2.]),
            "q": numpy.array([-0.5, 0.7])
        },
        doc="""Expected ranges of the state variables for initial condition generation and phase plane setup.""",
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={
            "r": numpy.array([0.0, numpy.inf])
        },
    )

    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=("r", "V", "g", "q"),
        default=("r", "V"),
        doc="The quantities of interest for monitoring for the Infinite QIF 2D oscillator.",
    )

    state_variables = ('r', 'V', 'g', 'q')
    _nvar = 4
    # Cvar is the coupling variable. 
    cvar = numpy.array([0, 1, 2, 3], dtype=numpy.int32)

    def dfun(self, state_variables, coupling, local_coupling=0.0):
        r"""
            4D model describing the Ott-Antonsen reduction of infinite all-to-all
            coupled QIF neurons (Theta-neurons) as in [Coombes_Byrne_2019]_.

            The equations of the model read

            .. math::
                    \dot{r} &= \Delta/\pi + 2 V r - g r^2 \\
                    \dot{V} &= V^2 - \pi^2 r^2 + \eta + (v_{syn} - V) g \\
                    \dot{g} &= \alpha q  \\
                    \dot{q} &= \alpha (\kappa \pi r - g - 2 q)
        """
        r, V, g, q = state_variables

        # [State_variables, nodes]
        Delta = self.Delta
        k = self.k
        v_syn = self.v_syn
        eta = self.eta
        alpha = self.alpha

        Coupling_Term_r = coupling[0, :]  # This zero refers to the first element of cvar (r in this case)

        derivative = numpy.empty_like(state_variables)

        derivative[0] = Delta / numpy.pi + 2 * V * r - g * r
        derivative[1] = V ** 2 - numpy.pi ** 2 * r ** 2 + eta + (v_syn - V) * g + Coupling_Term_r
        derivative[2] = alpha * (q)
        derivative[3] = alpha * (k * numpy.pi * r - g - 2 * q)

        return derivative


class CoombesByrne2D(Model):
    r"""
    2D model describing the Ott-Antonsen reduction of infinite all-to-all coupled 
    QIF neurons (Theta-neurons) as in [Coombes_Byrne_2019]_.

    The two state variables :math:`r` and :math:`V` represent the average firing 
    rate and the average membrane potential of our QIF neurons. The conductance 
    :math:`g` is not dynamical and proportional to :math:`r`.

    The equations of the model read
    
    .. math::
            \dot{r} &= \Delta/\pi + 2 V r - g r^2\\
            \dot{V} &= V^2 - \pi^2 r^2 + \eta + (v_{syn} - V) g \\
            g &= \kappa \pi r
    .. [Coombes_Byrne_2019] Coombes, S., & Byrne, Á. (2019). Next generation neural mass models. In *Nonlinear Dynamics in Computational Neuroscience* (pp. 1-16). Springer, Cham.

    """

    # Define traited attributes for this model, these represent possible kwargs.

    Delta = NArray(
        label=r":math:`\Delta`",
        default=numpy.array([1.]),
        domain=Range(lo=0.0, hi=10.0, step=0.01),
        doc="""Half-width of heterogeneous noise distribution""",
    )

    v_syn = NArray(
        label=":math:`v_syn`",
        default=numpy.array([-4.0]),
        domain=Range(lo=-20.0, hi=0.0, step=0.01),
        doc="""QIF membrane reversal potential""",
    )

    k = NArray(
        label=":math:`k`",
        default=numpy.array([1.]),
        domain=Range(lo=0., hi=5.0, step=0.01),
        doc="""Local coupling strength""",
    )

    eta = NArray(
        label=r":math:`\eta`",
        default=numpy.array([2.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.0001),
        doc="""Constant parameter to scale the rate of feedback from the
            firing rate variable to itself""",
    )

    # Informational attribute, used for phase-plane and initial()
    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={"r": numpy.array([0., 2.0]),
                 "V": numpy.array([-2.0, 1.5])},
        doc="""Expected ranges of the state variables for initial condition generation and phase plane setup.""",
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={
            "r": numpy.array([0.0, numpy.inf])
        },
    )

    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=("r", "V"),
        default=("r", "V"),
        doc="The quantities of interest for monitoring for the Infinite QIF 2D oscillator.",
    )

    state_variables = ('r', 'V')
    _nvar = 2
    # Cvar is the coupling variable. 
    cvar = numpy.array([0, 1], dtype=numpy.int32)

    def dfun(self, state_variables, coupling, local_coupling=0.0):
        r"""
           2D model describing the Ott-Antonsen reduction of infinite all-to-all coupled
           QIF neurons (Theta-neurons) as in [Coombes_Byrne_2019]_.

           The two state variables :math:`r` and :math:`V` represent the average firing
           rate and the average membrane potential of our QIF neurons. The conductance
           :math:`g` is not dynamical and proportional to :math:`r`.

           The equations of the model read

           .. math::
                   \dot{r} &= \Delta/\pi + 2 V r - g r^2\\
                   \dot{V} &= V^2 - \pi^2 r^2 + \eta + (v_{syn} - V) g \\
                   g &= \kappa \pi r
        """
        r, V = state_variables

        # [State_variables, nodes]
        Delta = self.Delta
        k = self.k
        v_syn = self.v_syn
        eta = self.eta

        Coupling_Term_r = coupling[0, :]  # This zero refers to the first element of cvar (r in this case)

        derivative = numpy.empty_like(state_variables)

        derivative[0] = Delta / numpy.pi + 2 * V * r - k * numpy.pi * r ** 2
        derivative[1] = V ** 2 - numpy.pi ** 2 * r ** 2 + eta + (v_syn - V) * k * numpy.pi * r + Coupling_Term_r

        return derivative


class GastSchmidtKnosche_SD(Model):
    r"""
    4D model describing the Ott-Antonsen reduction of infinite all-to-all 
    coupled QIF neurons (Theta-neurons) with Synaptic Depression adaptation 
    mechanisms [Gastetal_2020]_.

    The two state variables :math:`r` and :math:`V` represent the average firing rate and 
    the average membrane potential of our QIF neurons.
    :math:`A` and :math:`B` are respectively the adaptation variable and its derivative.

    The equations of the infinite QIF 2D population model read
    
    .. math::
            \dot{r} &= 1/\tau (\Delta/(\pi \tau) + 2 V r)\\
            \dot{V} &= 1/\tau (V^2 - \tau^2 \pi^2 r^2 + \eta + J \tau r (1 - A) + I)\\ 
            \dot{A} &= 1/\tau_A (B)\\
            \dot{B} &= 1/\tau_A (-2 B - A + \alpha  r) \\

    .. [Gastetal_2020] Gast, R., Schmidt, H., & Knösche, T. R. (2020). A mean-field description of bursting dynamics in spiking neural networks with short-term adaptation. *Neural Computation*, 32(9), 1615-1634.
    """

    # Define traited attributes for this model, these represent possible kwargs.

    tau = NArray(
        label=r":math:`\tau`",
        default=numpy.array([1.0]),
        domain=Range(lo=0., hi=15.0, step=0.01),
        doc="""Characteristic time""",
    )

    tau_A = NArray(
        label=r":math:`\tau_A`",
        default=numpy.array([10.0]),
        domain=Range(lo=0., hi=15.0, step=0.01),
        doc="""Adaptation time scale""",
    )

    alpha = NArray(
        label=r":math:`\alpha`",
        default=numpy.array([0.5]),
        domain=Range(lo=0.0, hi=1.0, step=0.1),
        doc="""adaptation rate""",
    )

    I = NArray(
        label=":math:`I_{ext}`",
        default=numpy.array([0.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.01),
        doc="""External homogeneous current""",
    )

    Delta = NArray(
        label=r":math:`\Delta`",
        default=numpy.array([2.0]),
        domain=Range(lo=0.0, hi=10.0, step=0.01),
        doc="""Half-width of heterogeneous noise distribution""",
    )

    J = NArray(
        label=":math:`J`",
        default=numpy.array([21.2132]),
        domain=Range(lo=-25.0, hi=25.0, step=0.01),
        doc="""Synaptic weight""",
    )

    eta = NArray(
        label=r":math:`\eta`",
        default=numpy.array([-6.]),
        domain=Range(lo=-10.0, hi=10.0, step=0.0001),
        doc="""Mean of heterogeneous noise distribution""",
    )

    cr = NArray(
        label=":math:`cr`",
        default=numpy.array([1.]),
        domain=Range(lo=0., hi=1, step=0.1),
        doc="""It is the weight on Coupling through variable r.""",
    )

    cv = NArray(
        label=":math:`cv`",
        default=numpy.array([0.]),
        domain=Range(lo=0., hi=1, step=0.1),
        doc="""It is the weight on Coupling through variable V.""",
    )

    # Informational attribute, used for phase-plane and initial()
    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={"r": numpy.array([0.0, 4]),
                 "V": numpy.array([-3.0, 0.3]),
                 "A": numpy.array([0.0, 0.4]),
                 "B": numpy.array([-0.2, 0.3])},
        doc="""Expected ranges of the state variables for initial condition generation and phase plane setup.""",
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={
            "r": numpy.array([0.0, numpy.inf])
        },
    )

    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=("r", "V", "A", "B"),
        default=("r", "V"),
        doc="The quantities of interest for monitoring for the Infinite QIF 2D oscillator.",
    )

    state_variables = ('r', 'V', 'A', 'B')
    _nvar = 4
    # Cvar is the coupling variable. 
    cvar = numpy.array([0, 1, 2, 3], dtype=numpy.int32)

    def dfun(self, state_variables, coupling, local_coupling=0.0):
        r"""
            4D model describing the Ott-Antonsen reduction of infinite all-to-all
            coupled QIF neurons (Theta-neurons) with Synaptic Depression adaptation
            mechanisms [Gastetal_2020]_.

            The two state variables :math:`r` and :math:`V` represent the average firing rate and
            the average membrane potential of our QIF neurons.
            :math:`A` and :math:`B` are respectively the adaptation variable and its derivative.

            The equations of the infinite QIF 2D population model read

            .. math::
                    \dot{r} &= 1/\tau (\Delta/(\pi \tau) + 2 V r)\\
                    \dot{V} &= 1/\tau (V^2 - \tau^2 \pi^2 r^2 + \eta + J \tau r (1 - A) + I)\\
                    \dot{A} &= 1/\tau_A (B)\\
                    \dot{B} &= 1/\tau_A (-2 B - A + \alpha  r) \\
        """
        r, V, A, B = state_variables

        # [State_variables, nodes]
        I = self.I
        Delta = self.Delta
        eta = self.eta
        J = self.J
        alpha = self.alpha
        cr = self.cr
        cv = self.cv
        alpha = self.alpha
        tau_A = self.tau_A
        tau = self.tau

        Coupling_Term_r = coupling[0, :]  # This zero refers to the first element of cvar (r in this case)
        Coupling_Term_V = coupling[1, :]  # This one refers to the second element of cvar (V in this case)

        derivative = numpy.empty_like(state_variables)

        derivative[0] = 1 / tau * (Delta / (numpy.pi * tau) + 2 * V * r)
        derivative[1] = 1 / tau * (V ** 2 - numpy.pi ** 2 * tau ** 2 * r ** 2 + eta + J * tau * r * (
                    1 - A) + I + cr * Coupling_Term_r + cv * Coupling_Term_V)
        derivative[2] = 1 / tau_A * (B)
        derivative[3] = 1 / tau_A * (- 2 * B - A + alpha * r)

        return derivative


class GastSchmidtKnosche_SF(Model):
    r"""
    4D model describing the Ott-Antonsen reduction of infinite all-to-all coupled QIF neurons (Theta-neurons) with Spike-Frequency adaptation mechanisms [Gastetal_2020]_.

    The two state variables :math:`r` and :math:`V` represent the average firing rate and 
    the average membrane potential of our QIF neurons.
    :math:`A` and :math:`B` are respectively the adaptation variable and its derivative.

    The equations of the infinite QIF 2D population model read
    
    .. math::
            \dot{r} &= 1/\tau (\Delta/(\pi \tau) + 2 V r)\\
            \dot{V} &= 1/\tau (V^2 - \tau^2 \pi^2 r^2 + \eta + J \tau r - A + I)\\ 
            \dot{A} &= 1/\tau_A (B)\\
            \dot{B} &= 1/\tau_A (-2 B - A + \alpha r) \\
    .. [Gastetal_2020]	Gast, R., Schmidt, H., & Knösche, T. R. (2020). A mean-field description of bursting dynamics in spiking neural networks with short-term adaptation. Neural Computation, 32(9), 1615-1634.
    """

    # Define traited attributes for this model, these represent possible kwargs.

    tau = NArray(
        label=r":math:`\tau`",
        default=numpy.array([1.0]),
        domain=Range(lo=0., hi=15.0, step=0.01),
        doc="""Characteristic time""",
    )

    tau_A = NArray(
        label=r":math:`\tau_A`",
        default=numpy.array([10.0]),
        domain=Range(lo=0., hi=15.0, step=0.01),
        doc="""Adaptation time scale""",
    )

    alpha = NArray(
        label=r":math:`\alpha`",
        default=numpy.array([10.]),
        domain=Range(lo=0.0, hi=1.0, step=0.1),
        doc="""adaptation rate""",
    )

    I = NArray(
        label=":math:`I_{ext}`",
        default=numpy.array([0.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.01),
        doc="""External homogeneous current""",
    )

    Delta = NArray(
        label=r":math:`\Delta`",
        default=numpy.array([2.0]),
        domain=Range(lo=0.0, hi=10.0, step=0.01),
        doc="""Half-width of heterogeneous noise distribution""",
    )

    J = NArray(
        label=":math:`J`",
        default=numpy.array([21.2132]),
        domain=Range(lo=-25.0, hi=25.0, step=0.01),
        doc="""Synaptic weight""",
    )

    eta = NArray(
        label=r":math:`\eta`",
        default=numpy.array([1.]),
        domain=Range(lo=-10.0, hi=10.0, step=0.0001),
        doc="""Mean of heterogeneous noise distribution""",
    )

    cr = NArray(
        label=":math:`cr`",
        default=numpy.array([1.]),
        domain=Range(lo=0., hi=1, step=0.1),
        doc="""It is the weight on Coupling through variable r.""",
    )

    cv = NArray(
        label=":math:`cv`",
        default=numpy.array([0.]),
        domain=Range(lo=0., hi=1, step=0.1),
        doc="""It is the weight on Coupling through variable V.""",
    )

    # Informational attribute, used for phase-plane and initial()
    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={"r": numpy.array([0., 2.0]),
                 "V": numpy.array([-2.0, 1.5]),
                 "A": numpy.array([-1., 1.0]),
                 "B": numpy.array([-1.0, 1.0])},
        doc="""Expected ranges of the state variables for initial condition generation and phase plane setup.""",
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={
            "r": numpy.array([0.0, numpy.inf])
        },
    )

    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=("r", "V", "A", "B"),
        default=("r", "V"),
        doc="The quantities of interest for monitoring for the Infinite QIF 2D oscillator.",
    )

    state_variables = ('r', 'V', 'A', 'B')
    _nvar = 4
    # Cvar is the coupling variable. 
    cvar = numpy.array([0, 1, 2, 3], dtype=numpy.int32)

    def dfun(self, state_variables, coupling, local_coupling=0.0):
        r"""
            4D model describing the Ott-Antonsen reduction of infinite all-to-all coupled QIF neurons (Theta-neurons) with Spike-Frequency adaptation mechanisms [Gastetal_2020]_.

            The two state variables :math:`r` and :math:`V` represent the average firing rate and
            the average membrane potential of our QIF neurons.
            :math:`A` and :math:`B` are respectively the adaptation variable and its derivative.

            The equations of the infinite QIF 2D population model read

            .. math::
                    \dot{r} &= 1/\tau (\Delta/(\pi \tau) + 2 V r)\\
                    \dot{V} &= 1/\tau (V^2 - \tau^2 \pi^2 r^2 + \eta + J \tau r - A + I)\\
                    \dot{A} &= 1/\tau_A (B)\\
                    \dot{B} &= 1/\tau_A (-2 B - A + \alpha r) \\
        """
        r, V, A, B = state_variables

        # [State_variables, nodes]
        I = self.I
        Delta = self.Delta
        eta = self.eta
        J = self.J
        alpha = self.alpha
        cr = self.cr
        cv = self.cv
        alpha = self.alpha
        tau_A = self.tau_A
        tau = self.tau

        Coupling_Term_r = coupling[0, :]  # This zero refers to the first element of cvar (r in this case)
        Coupling_Term_V = coupling[1, :]  # This one refers to the second element of cvar (V in this case)

        derivative = numpy.empty_like(state_variables)

        derivative[0] = 1 / tau * (Delta / (numpy.pi * tau) + 2 * V * r)
        derivative[1] = 1 / tau * (
                    V ** 2 - numpy.pi ** 2 * tau ** 2 * r ** 2 + eta + J * tau * r + I - A + cr * Coupling_Term_r + cv * Coupling_Term_V)
        derivative[2] = 1 / tau_A * (B)
        derivative[3] = 1 / tau_A * (- 2 * B - A + alpha * r)

        return derivative


class DumontGutkin(Model):
    r"""
    8D model describing the Ott-Antonsen reduction of infinite all-to-all 
    coupled QIF Excitatory E and Inhibitory I Theta-neurons with local synaptic 
    dynamics [DumontGutkin2019]_.

    State variables :math:`r` and :math:`V` represent the average firing rate and 
    the average membrane potential of our QIF neurons. 
    The neural masses are coupled through the firing rate of :math:`E_i` population from node i-th into :math:`E_j` and :math:`I_j` subpopulations in node j-th.

    The equations of the excitatory infinite QIF 4D population model read (similar for inhibitory):
    
    .. math::
            \dot{r} &= 1/\tau (\Delta/(\pi \tau) + 2 V r)\\
            \dot{V} &= 1/\tau (V^2 + \eta + \gamma I - \tau^2 \pi^2 r^2 + \tau g - \tau s)\\
            \dot{g} &= 1/\tau_s (-g + J_ r)\\
            \dot{s} &= 1/\tau_s (-s) \\

    .. [DumontGutkin2019] Dumont, G., & Gutkin, B. (2019). Macroscopic phase resetting-curves determine oscillatory coherence and signal transfer in inter-coupled neural circuits. PLoS computational biology, 15(5), e1007019.
    """

    # Define traited attributes for this model, these represent possible kwargs.
    I_e = NArray(
        label=":math:`I_{ext_e}`",
        default=numpy.array([0.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.01),
        doc="""External homogeneous current on excitatory population""",
    )

    Delta_e = NArray(
        label=r":math:`\Delta_e`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.0, hi=10.0, step=0.01),
        doc="""Half-width of heterogeneous noise distribution over excitatory population""",
    )

    eta_e = NArray(
        label=r":math:`\eta_e`",
        default=numpy.array([-5.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.0001),
        doc="""Mean heterogeneous current on excitatory population""",
    )

    tau_e = NArray(
        label=r":math:`\tau_e`",
        default=numpy.array([10.0]),
        domain=Range(lo=0., hi=15.0, step=0.01),
        doc="""Characteristic time of excitatory population""",
    )

    I_i = NArray(
        label=":math:`I_{ext_i}`",
        default=numpy.array([0.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.01),
        doc="""External current on inhibitory population""",
    )

    Delta_i = NArray(
        label=r":math:`\Delta_i`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.0, hi=10.0, step=0.01),
        doc="""Half-width of heterogeneous noise distribution over inhibitory population""",
    )

    eta_i = NArray(
        label=r":math:`\eta_i`",
        default=numpy.array([-5.0]),
        domain=Range(lo=-10.0, hi=10.0, step=0.0001),
        doc="""Mean heterogeneous current on inhibitory population""",
    )
    tau_i = NArray(
        label=r":math:`\tau_i`",
        default=numpy.array([10.0]),
        domain=Range(lo=0., hi=15.0, step=0.01),
        doc="""Characteristic time of inhibitory population""",
    )

    tau_s = NArray(
        label=r":math:`\tau_s`",
        default=numpy.array([1.0]),
        domain=Range(lo=0.0, hi=15.0, step=0.01),
        doc="""Synaptic time constant""",
    )

    J_ee = NArray(
        label=":math:`J_{ee}`",
        default=numpy.array([0.0]),
        domain=Range(lo=-25.0, hi=25.0, step=0.0001),
        doc="""Synaptic weight e-->e""",
    )

    J_ei = NArray(
        label=":math:`J_{ei}`",
        default=numpy.array([10.0]),
        domain=Range(lo=-25.0, hi=25.0, step=0.0001),
        doc="""Synaptic weight i-->e""",
    )

    J_ie = NArray(
        label=":math:`J_{ie}`",
        default=numpy.array([0.0]),
        domain=Range(lo=-25.0, hi=25.0, step=0.0001),
        doc="""Synaptic weight e-->i""",
    )

    J_ii = NArray(
        label=":math:`J_{ii}`",
        default=numpy.array([15.0]),
        domain=Range(lo=-25.0, hi=25.0, step=0.0001),
        doc="""Synaptic weight i-->i""",
    )

    Gamma = NArray(
        label=r":math:`\Gamma`",
        default=numpy.array([5.0]),
        domain=Range(lo=0., hi=10., step=0.1),
        doc="""Ratio of excitatory VS inhibitory global couplings G_ie/G_ee .""",
    )

    # Informational attribute, used for phase-plane and initial()
    state_variable_range = Final(
        label="State Variable ranges [lo, hi]",
        default={"r_e": numpy.array([0., 2.0]),
                 "V_e": numpy.array([-2.0, 1.5]),
                 "s_ee": numpy.array([-1.0, 1.0]),
                 "s_ei": numpy.array([-1.0, 1.0]),
                 "r_i": numpy.array([0., 2.0]),
                 "V_i": numpy.array([-2.0, 1.5]),
                 "s_ie": numpy.array([-1.0, 1.0]),
                 "s_ii": numpy.array([-1.0, 1.0])},
        doc="""Expected ranges of the state variables for initial condition generation and phase plane setup.""",
    )

    state_variable_boundaries = Final(
        label="State Variable boundaries [lo, hi]",
        default={
            "r_e": numpy.array([0.0, numpy.inf]),
            "r_i": numpy.array([0.0, numpy.inf])
        },
    )

    variables_of_interest = List(
        of=str,
        label="Variables or quantities available to Monitors",
        choices=("r_e", "V_e", "s_ee", "s_ei", "r_i", "V_i", "s_ie", "s_ii"),
        default=("r_e", "V_e", "s_ee", "s_ei", "r_i", "V_i", "s_ie", "s_ii"),
        doc="The quantities of interest for monitoring for the Infinite QIF 2D oscillator.",
    )

    state_variables = ('r_e', 'V_e', 's_ee', 's_ei', 'r_i', 'V_i', 's_ie', 's_ii')
    _nvar = 8
    # Cvar is the coupling variable. 
    cvar = numpy.array([0, 1, 4, 5], dtype=numpy.int32)

    def dfun(self, state_variables, coupling, local_coupling=0.0):
        r"""
            8D model describing the Ott-Antonsen reduction of infinite all-to-all
            coupled QIF Excitatory E and Inhibitory I Theta-neurons with local synaptic
            dynamics [DumontGutkin2019]_.

            State variables :math:`r` and :math:`V` represent the average firing rate and
            the average membrane potential of our QIF neurons.
            The neural masses are coupled through the firing rate of :math:`E_i` population from node i-th into :math:`E_j` and :math:`I_j` subpopulations in node j-th.

            The equations of the excitatory infinite QIF 4D population model read (similar for inhibitory):

            .. math::
                    \dot{r} &= 1/\tau (\Delta/(\pi \tau) + 2 V r)\\
                    \dot{V} &= 1/\tau (V^2 + \eta + \gamma I - \tau^2 \pi^2 r^2 + \tau g - \tau s)\\
                    \dot{g} &= 1/\tau_s (-g + J_ r)\\
                    \dot{s} &= 1/\tau_s (-s) \\
        """
        r_e, V_e, s_ee, s_ei, r_i, V_i, s_ie, s_ii = state_variables

        # [State_variables, nodes]
        Delta_e = self.Delta_e
        Delta_i = self.Delta_i
        tau_e = self.tau_e
        tau_i = self.tau_i
        tau_s = self.tau_s
        eta_e = self.eta_e
        eta_i = self.eta_i
        J_ee = self.J_ee
        J_ii = self.J_ii
        J_ei = self.J_ei
        J_ie = self.J_ie
        I_e = self.I_e
        I_i = self.I_i
        Gamma = self.Gamma

        Coupling_Term = coupling[0, :]  # This zero refers to the first element of cvar (r_e in this case)

        derivative = numpy.empty_like(state_variables)

        derivative[0] = 1 / tau_e * (Delta_e / (numpy.pi * tau_e) + 2 * V_e * r_e)
        derivative[1] = 1 / tau_e * (
                    V_e ** 2 + eta_e - tau_e ** 2 * numpy.pi ** 2 * r_e ** 2 + tau_e * s_ee - tau_e * s_ei + I_e)
        derivative[2] = 1 / tau_s * (- s_ee + J_ee * r_e + Coupling_Term)
        derivative[3] = 1 / tau_s * (- s_ei + J_ei * r_i)
        derivative[4] = 1 / tau_i * (Delta_i / (numpy.pi * tau_i) + 2 * V_i * r_i)
        derivative[5] = 1 / tau_i * (
                    V_i ** 2 + eta_i - tau_i ** 2 * numpy.pi ** 2 * r_i ** 2 + tau_i * s_ie - tau_i * s_ii + I_i)
        derivative[6] = 1 / tau_s * (- s_ie + J_ie * r_e + Gamma * Coupling_Term)
        derivative[7] = 1 / tau_s * (- s_ii + J_ii * r_i)

        return derivative



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Defines a set of integration methods for both deterministic and stochastic
differential equations.

Using an integration step size ``dt`` from the following list:
    [0.244140625, 0.1220703125, 0.06103515625, 0.048828125, 0.0244140625, 0.01220703125, 0.009765625, 0.006103515625, 0.0048828125]

will be consistent with Monitor periods corresponding to any of [4096, 2048, 1024, 512, 256, 128] Hz

# TODO: error analysis

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>
.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>
.. moduleauthor:: Noelia Montejo <Noelia@tvb.invalid>

"""
import abc
import functools
import numpy
import scipy.integrate
from . import noise
from .common import get_logger, simple_gen_astr
from tvb.basic.neotraits.api import HasTraits, Attr, NArray, Float

LOG = get_logger(__name__)


class Integrator(HasTraits):
    """
    The Integrator class is a base class for the integration methods...

    .. [1] Kloeden and Platen, Springer 1995, *Numerical solution of stochastic
        differential equations.*

    .. [2] Riccardo Mannella, *Integration of Stochastic Differential Equations
        on a Computer*, Int J. of Modern Physics C 13(9): 1177--1194, 2002.

    .. [3] R. Mannella and V. Palleschi, *Fast and precise algorithm for 
        computer simulation of stochastic differential equations*, Phys. Rev. A
        40: 3381, 1989.

    """

    dt = Float(
        label="Integration-step size (ms)",
        default=0.01220703125,
        required=True,
        doc="""The step size used by the integration routine in ms. This
        should be chosen to be small enough for the integration to be
        numerically stable. It is also necessary to consider the desired sample
        period of the Monitors, as they are restricted to being integral
        multiples of this value. The default value is set such that all built-in
        models are numerically stable with there default parameters and because
        it is consitent with Monitors using sample periods corresponding to
        powers of 2 from 128 to 4096Hz."""
    )

    bounded_state_variable_indices = NArray(
        dtype=int,
        label="indices of the state variables to be bounded by the integrators "
              "within the boundaries in the boundaries' values array",
        required=False)

    state_variable_boundaries = NArray(
        label="The boundary values of the state variables",
        required=False)

    clamped_state_variable_indices = NArray(
        dtype=int,
        label="indices of the state variables to be clamped by the integrators "
              "to the values in the clamped_values array",
        required=False)

    clamped_state_variable_values = NArray(
        label="The values of the state variables which are clamped ",
        required=False)

    _bounded_integration_state_variable_indices = None
    _integration_state_variable_boundaries = None
    _clamped_integration_state_variable_indices = None
    _clamped_integration_state_variable_values = None

    @abc.abstractmethod
    def scheme(self, X, dfun, coupling, local_coupling, stimulus):
        """
        The scheme of integrator should take a state and provide the next
        state in time, e.g. for a differential equation, scheme should take
        :math:`X` and provide an appropriate :math:`X + dX` (dfun in the code).

        """

    def set_random_state(self, random_state):
        self.log.warning("random_state supplied for non-stochastic integration")

    def configure(self):
        # Set default configurations:
        self._clamped_integration_state_variable_indices = self.clamped_state_variable_indices
        self._clamped_integration_state_variable_values = self.clamped_state_variable_values
        self._bounded_integration_state_variable_indices = self.bounded_state_variable_indices
        self._integration_state_variable_boundaries = self.state_variable_boundaries
        super(Integrator, self).configure()

    def configure_boundaries(self, model):
        if model.state_variable_boundaries is not None:
            indices = []
            boundaries = []
            for sv, sv_bounds in model.state_variable_boundaries.items():
                indices.append(model.state_variables.index(sv))
                boundaries.append(sv_bounds)
            sort_inds = numpy.argsort(indices)
            self.bounded_state_variable_indices = numpy.array(indices)[sort_inds]
            self.state_variable_boundaries = numpy.array(boundaries).astype("float64")[sort_inds]
            self._bounded_integration_state_variable_indices = numpy.copy(self.bounded_state_variable_indices)
            self._integration_state_variable_boundaries = numpy.copy(self.state_variable_boundaries)

    def reconfigure_boundaries_and_clamping_for_integration_state_variables(self, model):
        integration_state_variable_indices = numpy.where(model.state_variable_mask)[0].tolist()
        if self.state_variable_boundaries is not None:
            # If there are any state_variable_boundaries...
            bounded_integration_state_variable_indices = []
            integration_state_variable_boundaries = []
            # ...for each one of the bounded state variable indices and boundary values...
            for bound_sv_ind, bounds in zip(self._bounded_integration_state_variable_indices,
                                            self.state_variable_boundaries):
                # ...if the boundary indice corresponds to an integrated state variable...
                if bound_sv_ind in integration_state_variable_indices:
                    # ...add its integration state vector indice...
                    bounded_integration_state_variable_indices.append(
                        integration_state_variable_indices.index(bound_sv_ind))
                    # ...and the corresponding boundaries
                    integration_state_variable_boundaries.append(bounds)
            self._bounded_integration_state_variable_indices = \
                numpy.array(bounded_integration_state_variable_indices)
            self._integration_state_variable_boundaries = \
                numpy.array(integration_state_variable_boundaries)
        if self.clamped_state_variable_values is not None:
            # If there are any clamped values...
            clamped_integration_state_variable_indices = []
            clamped_integration_state_variable_values = []
            # ...for each one of the clamped state variable indices and clamped values...
            for clamp_sv_ind, clampval in zip(self.clamped_state_variable_indices,
                                              self.clamped_state_variable_values):
                # ...if the clamped indice corresponds to an integrated state variable...
                if clamp_sv_ind in integration_state_variable_indices:
                    # ...add its integration state vector indice...
                    clamped_integration_state_variable_indices.append(
                        integration_state_variable_indices.index(clamp_sv_ind))
                    # ...and the corresponding clamped value
                    clamped_integration_state_variable_values.append(clampval)
            self._clamped_integration_state_variable_indices = \
                numpy.array(clamped_integration_state_variable_indices)
            self._clamped_integration_state_variable_values = \
                numpy.array(clamped_integration_state_variable_values)

    def _bound_state(self, X, indices, boundaries):
        for sv_ind, sv_bounds in zip(indices, boundaries):
            if sv_bounds[0] is not None:
                X[sv_ind][X[sv_ind] < sv_bounds[0]] = sv_bounds[0]
            if sv_bounds[1] is not None:
                X[sv_ind][X[sv_ind] > sv_bounds[1]] = sv_bounds[1]

    def bound_state(self, X):
        self._bound_state(X, self.bounded_state_variable_indices, self.state_variable_boundaries)

    def bound_integration_state(self, X):
        self._bound_state(X, self._bounded_integration_state_variable_indices,
                          self._integration_state_variable_boundaries)

    def clamp_state(self, X):
        X[self.clamped_state_variable_indices] = self.clamped_state_variable_values

    def clamp_integration_state(self, X):
        X[self._clamped_integration_state_variable_indices] = self._clamped_integration_state_variable_values

    def bound_and_clamp(self, state):
        # If there is a state boundary...
        if self.state_variable_boundaries is not None:
            # ...use the integrator's bound_state
            self.bound_state(state)
        # If there is a state clamping...
        if self.clamped_state_variable_values is not None:
            # ...use the integrator's clamp_state
            self.clamp_state(state)

    def integration_bound_and_clamp(self, state):
        # If there is a state boundary...
        if self._integration_state_variable_boundaries is not None:
            # ...use the integrator's bound_state
            self.bound_integration_state(state)
        # If there is a state clamping...
        if self._clamped_integration_state_variable_values is not None:
            # ...use the integrator's clamp_state
            self.clamp_integration_state(state)

    def integrate_with_update(self, X, model, coupling, local_coupling, stimulus):
        temp = model.update_state_variables_before_integration(X, coupling, local_coupling, stimulus)
        if temp is not None:
            X = temp
            self.bound_and_clamp(X)
        X = self.integrate(X, model, coupling, local_coupling, stimulus)
        temp = model.update_state_variables_after_integration(X)
        if temp is not None:
            X = temp
            self.bound_and_clamp(X)
        return X

    def integrate(self, X, model, coupling, local_coupling, stimulus):
        X[model.state_variable_mask] = self.scheme(X[model.state_variable_mask],
                                                   model.dfun, coupling, local_coupling, stimulus)
        return X

    def __str__(self):
        return simple_gen_astr(self, 'dt')


class IntegratorStochastic(Integrator):
    r"""
    The IntegratorStochastic class is a base class for the stochastic
    integration methods. It derives from the Integrator abstract base class.

    We consider a stochastic differential equation has the generic form:

        .. math::
            \dot{X}_i(t) = dX_i(\vec{X}) + g_i(\vec{X})  \xi(t)

    where we assume that the stochastic process :math:`\xi` is a Gaussian
    forcing. In the deterministic case, :math:`g(X)` would be zero. The full
    algorithm, for one external stochastic forcing which is additive, is:

        .. math::
            X_i(t) = X_i(0) + g_i(X) Z_1(t) + dX_i t + Z_2(t) dX_{i,k} g_k(X) +
                    0.5 dX_{i,jk} g_j(X) g_k(X) Z_3(t) + 0.5 t^2 dX_{i,j} dX_j

    where :math:`Z_1`, :math:`Z_2` and :math:`Z_3` are Gaussian variables,
    assuming the Einstein notation and defining:

        .. math::
            dX_{i,j} = \frac{\partial dX_i}{\partial X_j}

    """

    noise = Attr(
        field_type=noise.Noise,
        label="Integration Noise",
        default=noise.Additive,
        required=True,
        doc="""The stochastic integrator's noise source. It incorporates its
        own instance of Numpy's RandomState.""")

    def set_random_state(self, random_state):
        if random_state is not None:
            self.noise.random_stream.set_state(random_state)
            msg = "random_state supplied with seed %s"
            self.log.info(msg, self.noise.random_stream.get_state()[1][0])

    def __str__(self):
        return simple_gen_astr(self, 'dt noise')


class HeunDeterministic(Integrator):
    """
    It is a simple example of a predictor-corrector method. It is also known as
    modified trapezoidal method, which uses the Euler method as its predictor.
    And it is also a implicit integration scheme.

    """

    n_dx = 2

    def scheme(self, X, dfun, coupling, local_coupling, stimulus):
        r"""
        From [1]_:

        .. math::
            X_{n+1} &= X_n + dt (dX(t_n, X_n) + 
                                 dX(t_{n+1}, \tilde{X}_{n+1})) / 2 \\
            \tilde{X}_{n+1} &= X_n + dt dX(t_n, X_n)

        cf. Equation 1.11, page 283.

        """
        #import pdb; pdb.set_trace()
        m_dx_tn = dfun(X, coupling, local_coupling)
        inter = X + self.dt * (m_dx_tn + stimulus)
        self.integration_bound_and_clamp(inter)

        dX = (m_dx_tn + dfun(inter, coupling, local_coupling)) * self.dt / 2.0

        X_next = X + dX + self.dt * stimulus
        self.integration_bound_and_clamp(X_next)

        return X_next


class HeunStochastic(IntegratorStochastic):
    """
    It is a simple example of a predictor-corrector method. It is also known as
    modified trapezoidal method, which uses the Euler method as its predictor.

    """

    n_dx = 2

    def scheme(self, X, dfun, coupling, local_coupling, stimulus):
        """
        From [2]_:

        .. math::
            X_i(t) = X_i(t-1) + dX(X_i(t)/2 + dX(X_i(t-1))) dt + g_i(X) Z_1

        in our case, :math:`noise = Z_1`

        See page 1180.

        """
        noise = self.noise.generate(X.shape)
        noise_gfun = self.noise.gfun(X)
        if (noise_gfun.shape != (1,) and noise.shape[0] != noise_gfun.shape[0]):
            msg = str("Got shape %s for noise but require %s."
                      " You need to reconfigure noise after you have changed your model."%(
                       noise_gfun.shape, (noise.shape[0], noise.shape[1])))
            raise Exception(msg)

        m_dx_tn = dfun(X, coupling, local_coupling)

        noise *= noise_gfun

        inter = X + self.dt * m_dx_tn + noise + self.dt * stimulus
        self.integration_bound_and_clamp(inter)

        dX = (m_dx_tn + dfun(inter, coupling, local_coupling)) * self.dt / 2.0

        X_next = X + dX + noise + self.dt * stimulus
        self.integration_bound_and_clamp(X_next)

        return X_next


class EulerDeterministic(Integrator):
    """
    It is the simplest difference methods for the initial value problem. The
    recursive structure of Euler scheme, which evaluates approximate values to
    the Ito process at the discretization instants only, is the key to its
    successful implementation.

    """

    n_dx = 1

    def scheme(self, X, dfun, coupling, local_coupling, stimulus):
        r"""

        .. math::
            X_{n+1} = X_n + dt \, dX(t_n, X_n)

        cf. Equations 1.3 and 1.13, pages 305 and 306 respectively.

        """

        self.dX = dfun(X, coupling, local_coupling) 

        X_next = X + self.dt * (self.dX + stimulus)
        self.integration_bound_and_clamp(X_next)

        return X_next


class EulerStochastic(IntegratorStochastic):
    """
    It is the simplest difference methods for the initial value problem. The
    recursive structure of Euler scheme, which evaluates approximate values to
    the Ito process at the discretization instants only, is the key to its
    successful implementation.

    """

    n_dx = 1

    def scheme(self, X, dfun, coupling, local_coupling, stimulus):
        r"""
        Ones of the simplest time discrete approximations of an Ito process is
        Euler-Maruyama approximation, that satisfies the scalar stochastic
        differential equation (From [1]_, page 305):

        .. math::
            X_{n+1} = X_n + dX(t_n, X_n) \, dt + g(X_n) Z_1

        in our case, :math:`noise = Z_1`

        cf. Equations 1.3 and 1.13, pages 305 and 306 respectively.

        """

        noise = self.noise.generate(X.shape)
        dX = dfun(X, coupling, local_coupling) * self.dt 
        noise_gfun = self.noise.gfun(X)
        X_next = X + dX + noise_gfun * noise + self.dt * stimulus
        self.integration_bound_and_clamp(X_next)

        return X_next


class RungeKutta4thOrderDeterministic(Integrator):
    """
    The Runge-Kutta method is a standard procedure with most one-step methods.

    """

    n_dx = 4

    def scheme(self, X, dfun, coupling, local_coupling=0.0, stimulus=0.0):
        r"""
        The classical 4th order Runge-Kutta method is an explicit method.
        The 4th order Runge-Kutta methods are the most commonly used,
        representing a good compromise between accuracy and computational
        effort.

        From [1]_, pages 289-290, cf. Equation 2.8

        .. math::
            y_{n+1} &= y_n + 1/6 h (k_1 + 2 k_2 + 2 k_3 + k_4) \\
            t_{n+1} &= t_n + h \\
            k_1 &= f(t_n, y_n) \\
            k_2 &= f(t_n + h/2, y_n + h k_1 / 2) \\
            k_3 &= f(t_n + h/2, y_n + h k_2 / 2) \\
            k_4 &= f(t_n + h, y_n + h k_3)


        """

        dt = self.dt
        dt2 = dt / 2.0
        dt6 = dt / 6.0

        k1 = dfun(X, coupling, local_coupling)
        inter_k1 = X + dt2 * k1
        self.integration_bound_and_clamp(inter_k1)

        k2 = dfun(inter_k1, coupling, local_coupling)
        inter_k2 = X + dt2 * k2
        self.integration_bound_and_clamp(inter_k2)

        k3 = dfun(inter_k2, coupling, local_coupling)
        inter_k3 = X + dt * k3
        self.integration_bound_and_clamp(inter_k3)

        k4 = dfun(inter_k3, coupling, local_coupling)

        dX = dt6 * (k1 + 2.0 * k2 + 2.0 * k3 + k4)

        X_next = X + dX + self.dt * stimulus
        self.integration_bound_and_clamp(X_next)

        return X_next


class Identity(Integrator):
    """
    The Identity integrator does not apply any scheme to the
    provided dfun, only returning its results.

    This allows the model to determine its stepping scheme
    directly, and may be used for difference equations and
    cellular automata.

    """

    n_dx = 1

    def scheme(self, X, dfun, coupling=None, local_coupling=0.0, stimulus=0.0):
        """
        The identity scheme simply returns the results of the dfun and
        stimulus.

        .. math::
            x_{n+1} = f(x_{n})

        """

        X_next = dfun(X, coupling, local_coupling) + stimulus
        self.integration_bound_and_clamp(X_next)
        return X_next


class IdentityStochastic(IntegratorStochastic):
    """
    A stochastic variant of the Identity integrator.  Together
    with time delays, this allows for MVAR models. 
    """

    n_dx = 1

    def scheme(self, X, dfun, coupling=None, local_coupling=0.0, stimulus=0.0):
        """
        The stochastic identity scheme simply returns the results of the dfun and
        the gfun and stimulus.

        .. math::
            x_{n+1} = f(x_{n}) + g(X_n) Z_1

        """
        z = self.noise.generate(X.shape) * self.noise.gfun(X)
        X_next = dfun(X, coupling, local_coupling) + z + stimulus
        self.integration_bound_and_clamp(X_next)
        return X_next


class SciPyODEBase(object):
    "Provides a base class for integrators using SciPy's ode class."

    def _dfun_wrapper(self, dfun, state_shape):
        @functools.wraps(dfun)
        def wrapper(t, X_, coupling=None, local_coupling=0.0):
            X = X_.reshape(state_shape)
            dXdt = dfun(X, coupling, local_coupling)
            return dXdt.ravel()
        return wrapper

    def _prepare_ode(self, X, dfun):
        ode = scipy.integrate.ode(self._dfun_wrapper(dfun, X.shape))
        ode.set_initial_value(X.ravel())
        ode.set_integrator(self._scipy_ode_integrator_name,
                           first_step=self.dt / 5.0)
        LOG.debug('prepped backing SciPy ODE integrator %r', ode)
        return ode

    _ode = None

    def _apply_ode(self, X, dfun, coupling, local_coupling, stimulus):
        if self._ode is None:
            self._ode = self._prepare_ode(X, dfun)
        self._ode.y[:] = X.ravel()
        self._ode.set_f_params(coupling, local_coupling)
        return self._ode.integrate(self._ode.t + self.dt).reshape(X.shape) + self.dt * stimulus


# TODO: Find a solution for boundary application for intermediate steps of SciPy O/SDE solvers
# Right now they would behave differently than the TVB ones.


class SciPyODE(SciPyODEBase):

    def scheme(self, X, dfun, coupling, local_coupling, stimulus):
        X_next = self._apply_ode(X, dfun, coupling, local_coupling, stimulus)
        self.integration_bound_and_clamp(X_next)
        return X_next

class SciPySDE(SciPyODEBase):

    def scheme(self, X, dfun, coupling, local_coupling, stimulus):
        X_next = self._apply_ode(X, dfun, coupling, local_coupling, stimulus)
        X_next += self.noise.gfun(X) * self.noise.generate(X.shape)
        self.integration_bound_and_clamp(X_next)
        return X_next


class VODE(SciPyODE, Integrator):
    _scipy_ode_integrator_name = "vode"


class VODEStochastic(SciPySDE, IntegratorStochastic):
    _scipy_ode_integrator_name = "vode"


class Dopri5(SciPyODE, Integrator):
    _scipy_ode_integrator_name = "dopri5"


class Dopri5Stochastic(SciPySDE, IntegratorStochastic):
    _scipy_ode_integrator_name = "dopri5"


class Dop853(SciPyODE, Integrator):
    _scipy_ode_integrator_name = "dop853"


class Dop853Stochastic(SciPySDE, IntegratorStochastic):
    _scipy_ode_integrator_name = "dop853"
